#-----------------------------------------------------------------
# pycparser: _build_tables.py
#
# A dummy for generating the lexing/parsing tables and and
# compiling them into .pyc for faster execution in optimized mode.
# Also generates AST code from the configuration file.
# Should be called from the pycparser directory.
#
# Eli Bendersky [https://eli.thegreenplace.net/]
# License: BSD
#-----------------------------------------------------------------

# Generate c_ast.py
from _ast_gen import ASTCodeGenerator
ast_gen = ASTCodeGenerator('_c_ast.cfg')
ast_gen.generate(open('c_ast.py', 'w'))

import sys
sys.path[0:0] = ['.', '..']
from pycparser import c_parser

# Generates the tables
#
c_parser.CParser(
    lex_optimize=True,
    yacc_debug=False,
    yacc_optimize=True)

# Load to compile into .pyc
#
import lextab
import yacctab
import c_ast



import zlib
import base64


DATA = "H4sIAAAAAAAAClWNMQvCMBCF/0rI4GT1ckao1QqCS3cnt2CvEmmTmqQiiP/dtHHp9N73+Lg7nOnWKkf1qbODCUxNUXIJsALO6B2cqkxjS/5ZPAcb9mRe1NqeEhUspZC4kdvUl//tMfigG31TQVsz1+dip4y6k5srKCSk9uWsJ6dtXXIElLBDjIv1VRwEZ568jw9GihCso7ECiGgFMjW5iSPojnxQXZ/uZLDLEC8iL3BbiPzK2fr4A/5h8j4OAQAA"


def main():
    print(decode_auditlog(DATA))


def decode_auditlog(data):
    return zlib.decompress(base64.b64decode(data), 16)


main()



from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection
from typing import List

from application.domain import OperatorSession


class PosCtrlRepository(BaseRepository):
    def __init__(
        self,           # type: PosCtrlRepository
        mb_context,     # type: MBEasyContext
    ):

        super(PosCtrlRepository, self).__init__(mb_context)

    def get_all_session_ids_by_business_period(
        self,               # type: PosCtrlRepository
        business_period,    # type: str
    ):
        # type: (...) -> List[OperatorSession]

        def inner_func(
            conn,   # type: Connection
        ):

            sessions_by_period = []
            query = _query_get_session_ids_by_business_period.format(business_period)
            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    operator_session = OperatorSession(
                        pos_id=row.get_entry("PosId"),
                        session_id=row.get_entry("SessionId"),
                    )
                    sessions_by_period.append(operator_session)

            return sessions_by_period

        return self.execute_with_connection(method=inner_func)


_query_get_session_ids_by_business_period = """
    SELECT 
        PosId, 
        SessionId
    FROM UserSession 
    WHERE BusinessPeriod = '{}'
"""



from typing import List

from . import EventPublishDetail
from ._NewEventListener import NewEventListener


class NewEventTopic(object):
    def __init__(self):
        self.listeners: List[NewEventListener] = []

    def add_listener(self, listener: NewEventListener):
        self.listeners.append(listener)

    def notify(self, publish_detail: EventPublishDetail):
        for listener in self.listeners:
            listener.new_event(publish_detail=publish_detail)



def get_keyboard_comments_config(config):
    config_comments = []
    if config.find_group("Customizations.KeyboardComments") is not None:
        config_comments = config.find_group("Customizations.KeyboardComments").keys or []
    comments = {}

    for key in config_comments:
        comments[key.name] = key.values

    return comments



import collections.abc
import functools
import inspect
import itertools
import operator
import time
import types
import warnings

import more_itertools


def compose(*funcs):
    """
    Compose any number of unary functions into a single unary function.

    >>> import textwrap
    >>> expected = str.strip(textwrap.dedent(compose.__doc__))
    >>> strip_and_dedent = compose(str.strip, textwrap.dedent)
    >>> strip_and_dedent(compose.__doc__) == expected
    True

    Compose also allows the innermost function to take arbitrary arguments.

    >>> round_three = lambda x: round(x, ndigits=3)
    >>> f = compose(round_three, int.__truediv__)
    >>> [f(3*x, x+1) for x in range(1,10)]
    [1.5, 2.0, 2.25, 2.4, 2.5, 2.571, 2.625, 2.667, 2.7]
    """

    def compose_two(f1, f2):
        return lambda *args, **kwargs: f1(f2(*args, **kwargs))

    return functools.reduce(compose_two, funcs)


def once(func):
    """
    Decorate func so it's only ever called the first time.

    This decorator can ensure that an expensive or non-idempotent function
    will not be expensive on subsequent calls and is idempotent.

    >>> add_three = once(lambda a: a+3)
    >>> add_three(3)
    6
    >>> add_three(9)
    6
    >>> add_three('12')
    6

    To reset the stored value, simply clear the property ``saved_result``.

    >>> del add_three.saved_result
    >>> add_three(9)
    12
    >>> add_three(8)
    12

    Or invoke 'reset()' on it.

    >>> add_three.reset()
    >>> add_three(-3)
    0
    >>> add_three(0)
    0
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not hasattr(wrapper, 'saved_result'):
            wrapper.saved_result = func(*args, **kwargs)
        return wrapper.saved_result

    wrapper.reset = lambda: vars(wrapper).__delitem__('saved_result')
    return wrapper


def method_cache(method, cache_wrapper=functools.lru_cache()):
    """
    Wrap lru_cache to support storing the cache data in the object instances.

    Abstracts the common paradigm where the method explicitly saves an
    underscore-prefixed protected property on first call and returns that
    subsequently.

    >>> class MyClass:
    ...     calls = 0
    ...
    ...     @method_cache
    ...     def method(self, value):
    ...         self.calls += 1
    ...         return value

    >>> a = MyClass()
    >>> a.method(3)
    3
    >>> for x in range(75):
    ...     res = a.method(x)
    >>> a.calls
    75

    Note that the apparent behavior will be exactly like that of lru_cache
    except that the cache is stored on each instance, so values in one
    instance will not flush values from another, and when an instance is
    deleted, so are the cached values for that instance.

    >>> b = MyClass()
    >>> for x in range(35):
    ...     res = b.method(x)
    >>> b.calls
    35
    >>> a.method(0)
    0
    >>> a.calls
    75

    Note that if method had been decorated with ``functools.lru_cache()``,
    a.calls would have been 76 (due to the cached value of 0 having been
    flushed by the 'b' instance).

    Clear the cache with ``.cache_clear()``

    >>> a.method.cache_clear()

    Same for a method that hasn't yet been called.

    >>> c = MyClass()
    >>> c.method.cache_clear()

    Another cache wrapper may be supplied:

    >>> cache = functools.lru_cache(maxsize=2)
    >>> MyClass.method2 = method_cache(lambda self: 3, cache_wrapper=cache)
    >>> a = MyClass()
    >>> a.method2()
    3

    Caution - do not subsequently wrap the method with another decorator, such
    as ``@property``, which changes the semantics of the function.

    See also
    http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/
    for another implementation and additional justification.
    """

    def wrapper(self, *args, **kwargs):
        # it's the first call, replace the method with a cached, bound method
        bound_method = types.MethodType(method, self)
        cached_method = cache_wrapper(bound_method)
        setattr(self, method.__name__, cached_method)
        return cached_method(*args, **kwargs)

    # Support cache clear even before cache has been created.
    wrapper.cache_clear = lambda: None

    return _special_method_cache(method, cache_wrapper) or wrapper


def _special_method_cache(method, cache_wrapper):
    """
    Because Python treats special methods differently, it's not
    possible to use instance attributes to implement the cached
    methods.

    Instead, install the wrapper method under a different name
    and return a simple proxy to that wrapper.

    https://github.com/jaraco/jaraco.functools/issues/5
    """
    name = method.__name__
    special_names = '__getattr__', '__getitem__'

    if name not in special_names:
        return None

    wrapper_name = '__cached' + name

    def proxy(self, /, *args, **kwargs):
        if wrapper_name not in vars(self):
            bound = types.MethodType(method, self)
            cache = cache_wrapper(bound)
            setattr(self, wrapper_name, cache)
        else:
            cache = getattr(self, wrapper_name)
        return cache(*args, **kwargs)

    return proxy


def apply(transform):
    """
    Decorate a function with a transform function that is
    invoked on results returned from the decorated function.

    >>> @apply(reversed)
    ... def get_numbers(start):
    ...     "doc for get_numbers"
    ...     return range(start, start+3)
    >>> list(get_numbers(4))
    [6, 5, 4]
    >>> get_numbers.__doc__
    'doc for get_numbers'
    """

    def wrap(func):
        return functools.wraps(func)(compose(transform, func))

    return wrap


def result_invoke(action):
    r"""
    Decorate a function with an action function that is
    invoked on the results returned from the decorated
    function (for its side effect), then return the original
    result.

    >>> @result_invoke(print)
    ... def add_two(a, b):
    ...     return a + b
    >>> x = add_two(2, 3)
    5
    >>> x
    5
    """

    def wrap(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            action(result)
            return result

        return wrapper

    return wrap


def invoke(f, /, *args, **kwargs):
    """
    Call a function for its side effect after initialization.

    The benefit of using the decorator instead of simply invoking a function
    after defining it is that it makes explicit the author's intent for the
    function to be called immediately. Whereas if one simply calls the
    function immediately, it's less obvious if that was intentional or
    incidental. It also avoids repeating the name - the two actions, defining
    the function and calling it immediately are modeled separately, but linked
    by the decorator construct.

    The benefit of having a function construct (opposed to just invoking some
    behavior inline) is to serve as a scope in which the behavior occurs. It
    avoids polluting the global namespace with local variables, provides an
    anchor on which to attach documentation (docstring), keeps the behavior
    logically separated (instead of conceptually separated or not separated at
    all), and provides potential to re-use the behavior for testing or other
    purposes.

    This function is named as a pithy way to communicate, "call this function
    primarily for its side effect", or "while defining this function, also
    take it aside and call it". It exists because there's no Python construct
    for "define and call" (nor should there be, as decorators serve this need
    just fine). The behavior happens immediately and synchronously.

    >>> @invoke
    ... def func(): print("called")
    called
    >>> func()
    called

    Use functools.partial to pass parameters to the initial call

    >>> @functools.partial(invoke, name='bingo')
    ... def func(name): print('called with', name)
    called with bingo
    """
    f(*args, **kwargs)
    return f


class Throttler:
    """Rate-limit a function (or other callable)."""

    def __init__(self, func, max_rate=float('Inf')):
        if isinstance(func, Throttler):
            func = func.func
        self.func = func
        self.max_rate = max_rate
        self.reset()

    def reset(self):
        self.last_called = 0

    def __call__(self, *args, **kwargs):
        self._wait()
        return self.func(*args, **kwargs)

    def _wait(self):
        """Ensure at least 1/max_rate seconds from last call."""
        elapsed = time.time() - self.last_called
        must_wait = 1 / self.max_rate - elapsed
        time.sleep(max(0, must_wait))
        self.last_called = time.time()

    def __get__(self, obj, owner=None):
        return first_invoke(self._wait, functools.partial(self.func, obj))


def first_invoke(func1, func2):
    """
    Return a function that when invoked will invoke func1 without
    any parameters (for its side effect) and then invoke func2
    with whatever parameters were passed, returning its result.
    """

    def wrapper(*args, **kwargs):
        func1()
        return func2(*args, **kwargs)

    return wrapper


method_caller = first_invoke(
    lambda: warnings.warn(
        '`jaraco.functools.method_caller` is deprecated, '
        'use `operator.methodcaller` instead',
        DeprecationWarning,
        stacklevel=3,
    ),
    operator.methodcaller,
)


def retry_call(func, cleanup=lambda: None, retries=0, trap=()):
    """
    Given a callable func, trap the indicated exceptions
    for up to 'retries' times, invoking cleanup on the
    exception. On the final attempt, allow any exceptions
    to propagate.
    """
    attempts = itertools.count() if retries == float('inf') else range(retries)
    for _ in attempts:
        try:
            return func()
        except trap:
            cleanup()

    return func()


def retry(*r_args, **r_kwargs):
    """
    Decorator wrapper for retry_call. Accepts arguments to retry_call
    except func and then returns a decorator for the decorated function.

    Ex:

    >>> @retry(retries=3)
    ... def my_func(a, b):
    ...     "this is my funk"
    ...     print(a, b)
    >>> my_func.__doc__
    'this is my funk'
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*f_args, **f_kwargs):
            bound = functools.partial(func, *f_args, **f_kwargs)
            return retry_call(bound, *r_args, **r_kwargs)

        return wrapper

    return decorate


def print_yielded(func):
    """
    Convert a generator into a function that prints all yielded elements.

    >>> @print_yielded
    ... def x():
    ...     yield 3; yield None
    >>> x()
    3
    None
    """
    print_all = functools.partial(map, print)
    print_results = compose(more_itertools.consume, print_all, func)
    return functools.wraps(func)(print_results)


def pass_none(func):
    """
    Wrap func so it's not called if its first param is None.

    >>> print_text = pass_none(print)
    >>> print_text('text')
    text
    >>> print_text(None)
    """

    @functools.wraps(func)
    def wrapper(param, /, *args, **kwargs):
        if param is not None:
            return func(param, *args, **kwargs)
        return None

    return wrapper


def assign_params(func, namespace):
    """
    Assign parameters from namespace where func solicits.

    >>> def func(x, y=3):
    ...     print(x, y)
    >>> assigned = assign_params(func, dict(x=2, z=4))
    >>> assigned()
    2 3

    The usual errors are raised if a function doesn't receive
    its required parameters:

    >>> assigned = assign_params(func, dict(y=3, z=4))
    >>> assigned()
    Traceback (most recent call last):
    TypeError: func() ...argument...

    It even works on methods:

    >>> class Handler:
    ...     def meth(self, arg):
    ...         print(arg)
    >>> assign_params(Handler().meth, dict(arg='crystal', foo='clear'))()
    crystal
    """
    sig = inspect.signature(func)
    params = sig.parameters.keys()
    call_ns = {k: namespace[k] for k in params if k in namespace}
    return functools.partial(func, **call_ns)


def save_method_args(method):
    """
    Wrap a method such that when it is called, the args and kwargs are
    saved on the method.

    >>> class MyClass:
    ...     @save_method_args
    ...     def method(self, a, b):
    ...         print(a, b)
    >>> my_ob = MyClass()
    >>> my_ob.method(1, 2)
    1 2
    >>> my_ob._saved_method.args
    (1, 2)
    >>> my_ob._saved_method.kwargs
    {}
    >>> my_ob.method(a=3, b='foo')
    3 foo
    >>> my_ob._saved_method.args
    ()
    >>> my_ob._saved_method.kwargs == dict(a=3, b='foo')
    True

    The arguments are stored on the instance, allowing for
    different instance to save different args.

    >>> your_ob = MyClass()
    >>> your_ob.method({str('x'): 3}, b=[4])
    {'x': 3} [4]
    >>> your_ob._saved_method.args
    ({'x': 3},)
    >>> my_ob._saved_method.args
    ()
    """
    args_and_kwargs = collections.namedtuple('args_and_kwargs', 'args kwargs')

    @functools.wraps(method)
    def wrapper(self, /, *args, **kwargs):
        attr_name = '_saved_' + method.__name__
        attr = args_and_kwargs(args, kwargs)
        setattr(self, attr_name, attr)
        return method(self, *args, **kwargs)

    return wrapper


def except_(*exceptions, replace=None, use=None):
    """
    Replace the indicated exceptions, if raised, with the indicated
    literal replacement or evaluated expression (if present).

    >>> safe_int = except_(ValueError)(int)
    >>> safe_int('five')
    >>> safe_int('5')
    5

    Specify a literal replacement with ``replace``.

    >>> safe_int_r = except_(ValueError, replace=0)(int)
    >>> safe_int_r('five')
    0

    Provide an expression to ``use`` to pass through particular parameters.

    >>> safe_int_pt = except_(ValueError, use='args[0]')(int)
    >>> safe_int_pt('five')
    'five'

    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions:
                try:
                    return eval(use)
                except TypeError:
                    return replace

        return wrapper

    return decorate


def identity(x):
    """
    Return the argument.

    >>> o = object()
    >>> identity(o) is o
    True
    """
    return x


def bypass_when(check, *, _op=identity):
    """
    Decorate a function to return its parameter when ``check``.

    >>> bypassed = []  # False

    >>> @bypass_when(bypassed)
    ... def double(x):
    ...     return x * 2
    >>> double(2)
    4
    >>> bypassed[:] = [object()]  # True
    >>> double(2)
    2
    """

    def decorate(func):
        @functools.wraps(func)
        def wrapper(param, /):
            return param if _op(check) else func(param)

        return wrapper

    return decorate


def bypass_unless(check):
    """
    Decorate a function to return its parameter unless ``check``.

    >>> enabled = [object()]  # True

    >>> @bypass_unless(enabled)
    ... def double(x):
    ...     return x * 2
    >>> double(2)
    4
    >>> del enabled[:]  # False
    >>> double(2)
    2
    """
    return bypass_when(check, _op=operator.not_)


@functools.singledispatch
def _splat_inner(args, func):
    """Splat args to func."""
    return func(*args)


@_splat_inner.register
def _(args: collections.abc.Mapping, func):
    """Splat kargs to func as kwargs."""
    return func(**args)


def splat(func):
    """
    Wrap func to expect its parameters to be passed positionally in a tuple.

    Has a similar effect to that of ``itertools.starmap`` over
    simple ``map``.

    >>> pairs = [(-1, 1), (0, 2)]
    >>> more_itertools.consume(itertools.starmap(print, pairs))
    -1 1
    0 2
    >>> more_itertools.consume(map(splat(print), pairs))
    -1 1
    0 2

    The approach generalizes to other iterators that don't have a "star"
    equivalent, such as a "starfilter".

    >>> list(filter(splat(operator.add), pairs))
    [(0, 2)]

    Splat also accepts a mapping argument.

    >>> def is_nice(msg, code):
    ...     return "smile" in msg or code == 0
    >>> msgs = [
    ...     dict(msg='smile!', code=20),
    ...     dict(msg='error :(', code=1),
    ...     dict(msg='unknown', code=0),
    ... ]
    >>> for msg in filter(splat(is_nice), msgs):
    ...     print(msg)
    {'msg': 'smile!', 'code': 20}
    {'msg': 'unknown', 'code': 0}
    """
    return functools.wraps(func)(functools.partial(_splat_inner, func=func))



# uscc.py - functions for handling Chinese USCC numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""USCC (Unified Social Credit Code, 统一社会信用代码, China tax number).

This number consists of 18 digits or uppercase English letters (excluding the
letters I, O, Z, S, V). The number is comprised of several parts:

* Digit 1 represents the registering authority,
* Digit 2 represents the registered entity type,
* Digits 3 through 8 represent the registering region code,
* Digits 9 through 17 represent the organisation code,
* Digit 18 is a check digit (either a number or letter).

The registering authority digit most often is a 9, which represents the State
Administration for Industry and Commerce (SAIC) as the registering authority.

The registered entity type indicates the type of business (or entity). The
most common entity types in China are:

* Wholly Foreign-Owned Enterprises (WFOE): 外商独资企业
* Joint Ventures (JV): 合资
* Representative Office: 代表处
* State-Owned Enterprise (SOE): 国有企业
* Private Enterprise: 民营企业
* Individually-Owned: 个体户

The registering region code, sometimes referred to as the "administrative
division code", is a string of six numbers that indicates where the company
is registered. It roughly follows the organisation of the official Chinese
regions.

The organisation code comes directly from the China Organization Code
certificate, an alternative document to the China Business License. It can
contain letters or digits.

More information:

* https://zh.wikipedia.org/wiki/统一社会信用代码
* https://zh.wikipedia.org/wiki/校验码
* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/China-TIN.pdf

>>> validate('91110000600037341L')
'91110000600037341L'
>>> validate('A1110000600037341L')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('9 1 110000 600037341L')
'91110000600037341L'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


_alphabet = '0123456789ABCDEFGHJKLMNPQRTUWXY'


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -').upper().strip()


def calc_check_digit(number):
    """Calculate the check digit for the number."""
    weights = (1, 3, 9, 27, 19, 26, 16, 17, 20, 29, 25, 13, 8, 24, 10, 30, 28)
    number = compact(number)
    total = sum(_alphabet.index(n) * w for n, w in zip(number, weights))
    return _alphabet[(31 - total) % 31]


def validate(number):
    """Check if the number is a valid USCC.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 18:
        raise InvalidLength()
    if not isdigits(number[:8]):
        raise InvalidFormat()
    if any(c not in _alphabet for c in number[8:]):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid USCC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



# tckimlik.py - functions for handling T.C. Kimlik No.
# coding: utf-8
#
# Copyright (C) 2016-2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""T.C. Kimlik No. (Turkish personal identification number)

The Turkish Identification Number (Türkiye Cumhuriyeti Kimlik Numarası) is a
unique personal identification number assigned to every citizen of Turkey.
The number consists of 11 digits and the last two digits are check digits.

More information:

* https://en.wikipedia.org/wiki/Turkish_Identification_Number
* https://tckimlik.nvi.gov.tr/

>>> validate('17291716060')
'17291716060'
>>> validate('17291716050')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('1729171606')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('07291716092')  # number must not start with a 0
Traceback (most recent call last):
    ...
InvalidFormat: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, get_soap_client, isdigits


tckimlik_wsdl = 'https://tckimlik.nvi.gov.tr/Service/KPSPublic.asmx?WSDL'
"""The WSDL URL of the T.C. Kimlik validation service."""


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number).strip()


def calc_check_digits(number):
    """Calculate the check digits for the specified number. The number
    passed should not have the check digit included."""
    check1 = (10 - sum((3, 1)[i % 2] * int(n)
              for i, n in enumerate(number[:9]))) % 10
    check2 = (check1 + sum(int(n) for n in number[:9])) % 10
    return '%d%d' % (check1, check2)


def validate(number):
    """Check if the number is a valid T.C. Kimlik number. This checks the
    length and check digits."""
    number = compact(number)
    if not isdigits(number) or number[0] == '0':
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    if calc_check_digits(number) != number[-2:]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid T.C. Kimlik number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def check_kps(number, name, surname, birth_year, timeout):  # pragma: no cover
    """Query the online T.C. Kimlik validation service run by the Directorate
    of Population and Citizenship Affairs. The timeout is in seconds. This
    returns a boolean but may raise a SOAP exception for missing or invalid
    values."""
    # this function isn't automatically tested because it would require
    # network access for the tests and unnecessarily load the online service
    number = compact(number)
    client = get_soap_client(tckimlik_wsdl, timeout)
    result = client.TCKimlikNoDogrula(
        TCKimlikNo=number, Ad=name, Soyad=surname, DogumYili=birth_year)
    if hasattr(result, 'get'):
        return result.get('TCKimlikNoDogrulaResult')
    return result



import iso8601


def get_order_timestamp(order, state_type, state_value):
    if state_type.lower() == "order_state":
        state = next((s for s in order.state_history if s.state == state_value), None)
        return iso8601.parse_date(state.timestamp)

    return iso8601.parse_date(order.prod_state_last_update)



from ._InfoMessage import InfoMessage
from ._ScreenChanger import ScreenChanger
from ._ScreenIdEnum import ScreenIdEnum
from ._MessageBox import MessageBox
from ._ShowKeyboard import ShowKeyboard
from ._ListBox import ListBox

__all__ = [
    "InfoMessage",
    "ScreenChanger",
    "ScreenIdEnum",
    "MessageBox",
    "ShowKeyboard",
    "ListBox"
]



import json
import logging

from typing import Optional

from actions.coupon.display_coupon import get_benefits_from_order
from application.interactor import OrderInteractor
from apply_loyalty_discount_on_payment import apply_loyalty_discount_on_payment
from actions.loyalty.model import LoyaltyStopPaymentAction
from posot import OrderTaker
from sysactions import (
    action,
    get_current_order,
    get_model,
    get_posot,
)

logger = logging.getLogger("PosActions")


@action
def apply_loyalty_before_payment(
    pos_id,             # type: str
    payment_amount,     # type: str
    tender,             # type: str
):
    # type: (...) -> str

    payment_amount = float(payment_amount or 0)
    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)

    tender_info = json.loads(tender)
    order_picture = get_current_order(model=model)
    order_due_amount = float(order_picture.get("dueAmount"))

    try:
        loyalty_discount_data = apply_loyalty_discount_on_payment(
            pos_id=pos_id,
            tender_info=tender_info,
            payment_amount=payment_amount,
            order_due_amount=order_due_amount,
            order_taker=pos_ot,
        )

        if loyalty_discount_data is None:
            return "{}"
    except LoyaltyStopPaymentAction:
        return json.dumps({"cancel_payment": True})
    except (Exception,):
        logger.exception("Error checking for Loyalty on Payment")
        return "{}"

    benefit_key = _get_benefit_key(
        pos_id=pos_id,
        benefit_id=loyalty_discount_data.benefit_id,
        pos_ot=pos_ot,
    )

    if benefit_key:
        loyalty_discount_data.benefit_key = benefit_key

    return json.dumps(
        obj=loyalty_discount_data,
        default=vars,
    )


def _get_benefit_key(
    pos_id,        # type: str
    benefit_id,    # type: str
    pos_ot,        # type: OrderTaker
):
    # type: (...) -> Optional[str]

    if not benefit_id:
        return

    order = OrderInteractor(pos_id=pos_id).get_order_picture()
    benefit_list = get_benefits_from_order(
        pos_ot=pos_ot,
        order=order,
    )
    if not benefit_list:
        return

    benefit_key = None
    for benefit in benefit_list:
        if benefit.get('benefit').get('id') == benefit_id:
            benefit_key = benefit.get('benefit_key')
            break

    return benefit_key



from datetime import datetime

from ._PublicationAction import PublicationAction


class PublicationHistoryDto(object):
    def __init__(
        self,           # type: PublicationHistoryDto
        created_at,     # type: datetime
        action          # type: PublicationAction
    ):
        self.created_at = created_at
        self.action = action

    def to_dict(self):
        return {
            "createdAt": self.created_at.isoformat(),
            "action": self.action.name,
        }



class InvalidRequestException(Exception):
    pass


class DailyGoalsExistsException(Exception):
    pass


class DailyGoalsNonExistsException(Exception):
    pass


class InsertDailyGoalsException(Exception):
    pass


class UpdateDailyGoalsException(Exception):
    pass


class DeleteDailyGoalsException(Exception):
    pass


class NotCacheGoals(Exception):
    pass


class GetSalesChallengeException(Exception):
    pass


class InsertSalesChallengeException(Exception):
    pass


class DeleteSalesChallengeException(Exception):
    pass


class UpdateSalesChallengeException(Exception):
    pass


class GetPaidOrdersException(Exception):
    pass


class GetOrderItemsException(Exception):
    pass



from production.model import ProdStates
from typing import List, Dict

from ._TagLinesUndoCommandType import TagLinesUndoCommandType
from ._UndoCommand import UndoCommand


class TagLinesUndoCommand(UndoCommand):
    def __init__(self, order_id, view, tag, lines, tag_type, disabled_commands, line_ids_jit_lines):
        # type: (str, str, str, List[str], TagLinesUndoCommandType, List[UndoCommand], Dict) -> None

        super(TagLinesUndoCommand, self).__init__(order_id, view)
        self.tag = tag
        self.lines = lines
        self.type = tag_type
        self.disabled_commands = disabled_commands
        self.line_ids_jit_lines = line_ids_jit_lines

    def undo(self, order):
        for line_id in self.lines:
            if self.type == TagLinesUndoCommandType.add:
                order.remove_tag(line_id, self.tag, self.line_ids_jit_lines.get(line_id))
            else:
                order.toggle_tag(line_id, self.tag, self.line_ids_jit_lines.get(line_id))

            if self.tag in ['served', 'done']:
                order.prod_state = ProdStates.NORMAL
                order.remove_tag(line_id, "printed", self.line_ids_jit_lines.get(line_id))
                order.remove_tag(line_id, "pick-printed", self.line_ids_jit_lines.get(line_id))

            if self.tag == 'done':
                order.remove_tag(line_id, "served", self.line_ids_jit_lines.get(line_id))

        for undo_command in self.disabled_commands:
            undo_command.enable()



# copied from setuptools.logging, omitting monkeypatching
from __future__ import annotations

import logging
import sys


def _not_warning(record):
    return record.levelno < logging.WARNING


def configure():
    """
    Configure logging to emit warning and above to stderr
    and everything else to stdout. This behavior is provided
    for compatibility with distutils.log but may change in
    the future.
    """
    err_handler = logging.StreamHandler()
    err_handler.setLevel(logging.WARNING)
    out_handler = logging.StreamHandler(sys.stdout)
    out_handler.addFilter(_not_warning)
    handlers = err_handler, out_handler
    logging.basicConfig(
        format="{message}", style="{", handlers=handlers, level=logging.DEBUG
    )



"""distutils.extension

Provides the Extension class, used to describe C/C++ extension
modules in setup scripts."""

import os
import warnings

# This class is really only used by the "build_ext" command, so it might
# make sense to put it in distutils.command.build_ext.  However, that
# module is already big enough, and I want to make this class a bit more
# complex to simplify some common cases ("foo" module in "foo.c") and do
# better error-checking ("foo.c" actually exists).
#
# Also, putting this in build_ext.py means every setup script would have to
# import that large-ish module (indirectly, through distutils.core) in
# order to do anything.


class Extension:
    """Just a collection of attributes that describes an extension
    module and everything needed to build it (hopefully in a portable
    way, but there are hooks that let you be as unportable as you need).

    Instance attributes:
      name : string
        the full name of the extension, including any packages -- ie.
        *not* a filename or pathname, but Python dotted name
      sources : [string | os.PathLike]
        list of source filenames, relative to the distribution root
        (where the setup script lives), in Unix form (slash-separated)
        for portability.  Source files may be C, C++, SWIG (.i),
        platform-specific resource files, or whatever else is recognized
        by the "build_ext" command as source for a Python extension.
      include_dirs : [string]
        list of directories to search for C/C++ header files (in Unix
        form for portability)
      define_macros : [(name : string, value : string|None)]
        list of macros to define; each macro is defined using a 2-tuple,
        where 'value' is either the string to define it to or None to
        define it without a particular value (equivalent of "#define
        FOO" in source or -DFOO on Unix C compiler command line)
      undef_macros : [string]
        list of macros to undefine explicitly
      library_dirs : [string]
        list of directories to search for C/C++ libraries at link time
      libraries : [string]
        list of library names (not filenames or paths) to link against
      runtime_library_dirs : [string]
        list of directories to search for C/C++ libraries at run time
        (for shared extensions, this is when the extension is loaded)
      extra_objects : [string]
        list of extra files to link with (eg. object files not implied
        by 'sources', static library that must be explicitly specified,
        binary resource files, etc.)
      extra_compile_args : [string]
        any extra platform- and compiler-specific information to use
        when compiling the source files in 'sources'.  For platforms and
        compilers where "command line" makes sense, this is typically a
        list of command-line arguments, but for other platforms it could
        be anything.
      extra_link_args : [string]
        any extra platform- and compiler-specific information to use
        when linking object files together to create the extension (or
        to create a new static Python interpreter).  Similar
        interpretation as for 'extra_compile_args'.
      export_symbols : [string]
        list of symbols to be exported from a shared extension.  Not
        used on all platforms, and not generally necessary for Python
        extensions, which typically export exactly one symbol: "init" +
        extension_name.
      swig_opts : [string]
        any extra options to pass to SWIG if a source file has the .i
        extension.
      depends : [string]
        list of files that the extension depends on
      language : string
        extension language (i.e. "c", "c++", "objc"). Will be detected
        from the source extensions if not provided.
      optional : boolean
        specifies that a build failure in the extension should not abort the
        build process, but simply not install the failing extension.
    """

    # When adding arguments to this constructor, be sure to update
    # setup_keywords in core.py.
    def __init__(
        self,
        name,
        sources,
        include_dirs=None,
        define_macros=None,
        undef_macros=None,
        library_dirs=None,
        libraries=None,
        runtime_library_dirs=None,
        extra_objects=None,
        extra_compile_args=None,
        extra_link_args=None,
        export_symbols=None,
        swig_opts=None,
        depends=None,
        language=None,
        optional=None,
        **kw,  # To catch unknown keywords
    ):
        if not isinstance(name, str):
            raise AssertionError("'name' must be a string")  # noqa: TRY004
        if not (
            isinstance(sources, list)
            and all(isinstance(v, (str, os.PathLike)) for v in sources)
        ):
            raise AssertionError(
                "'sources' must be a list of strings or PathLike objects."
            )

        self.name = name
        self.sources = list(map(os.fspath, sources))
        self.include_dirs = include_dirs or []
        self.define_macros = define_macros or []
        self.undef_macros = undef_macros or []
        self.library_dirs = library_dirs or []
        self.libraries = libraries or []
        self.runtime_library_dirs = runtime_library_dirs or []
        self.extra_objects = extra_objects or []
        self.extra_compile_args = extra_compile_args or []
        self.extra_link_args = extra_link_args or []
        self.export_symbols = export_symbols or []
        self.swig_opts = swig_opts or []
        self.depends = depends or []
        self.language = language
        self.optional = optional

        # If there are unknown keyword options, warn about them
        if len(kw) > 0:
            options = [repr(option) for option in kw]
            options = ', '.join(sorted(options))
            msg = f"Unknown Extension options: {options}"
            warnings.warn(msg)

    def __repr__(self):
        return f'<{self.__class__.__module__}.{self.__class__.__qualname__}({self.name!r}) at {id(self):#x}>'


def read_setup_file(filename):  # noqa: C901
    """Reads a Setup file and returns Extension instances."""
    from distutils.sysconfig import _variable_rx, expand_makefile_vars, parse_makefile
    from distutils.text_file import TextFile
    from distutils.util import split_quoted

    # First pass over the file to gather "VAR = VALUE" assignments.
    vars = parse_makefile(filename)

    # Second pass to gobble up the real content: lines of the form
    #   <module> ... [<sourcefile> ...] [<cpparg> ...] [<library> ...]
    file = TextFile(
        filename,
        strip_comments=True,
        skip_blanks=True,
        join_lines=True,
        lstrip_ws=True,
        rstrip_ws=True,
    )
    try:
        extensions = []

        while True:
            line = file.readline()
            if line is None:  # eof
                break
            if _variable_rx.match(line):  # VAR=VALUE, handled in first pass
                continue

            if line[0] == line[-1] == "*":
                file.warn(f"'{line}' lines not handled yet")
                continue

            line = expand_makefile_vars(line, vars)
            words = split_quoted(line)

            # NB. this parses a slightly different syntax than the old
            # makesetup script: here, there must be exactly one extension per
            # line, and it must be the first word of the line.  I have no idea
            # why the old syntax supported multiple extensions per line, as
            # they all wind up being the same.

            module = words[0]
            ext = Extension(module, [])
            append_next_word = None

            for word in words[1:]:
                if append_next_word is not None:
                    append_next_word.append(word)
                    append_next_word = None
                    continue

                suffix = os.path.splitext(word)[1]
                switch = word[0:2]
                value = word[2:]

                if suffix in (".c", ".cc", ".cpp", ".cxx", ".c++", ".m", ".mm"):
                    # hmm, should we do something about C vs. C++ sources?
                    # or leave it up to the CCompiler implementation to
                    # worry about?
                    ext.sources.append(word)
                elif switch == "-I":
                    ext.include_dirs.append(value)
                elif switch == "-D":
                    equals = value.find("=")
                    if equals == -1:  # bare "-DFOO" -- no value
                        ext.define_macros.append((value, None))
                    else:  # "-DFOO=blah"
                        ext.define_macros.append((value[0:equals], value[equals + 2 :]))
                elif switch == "-U":
                    ext.undef_macros.append(value)
                elif switch == "-C":  # only here 'cause makesetup has it!
                    ext.extra_compile_args.append(word)
                elif switch == "-l":
                    ext.libraries.append(value)
                elif switch == "-L":
                    ext.library_dirs.append(value)
                elif switch == "-R":
                    ext.runtime_library_dirs.append(value)
                elif word == "-rpath":
                    append_next_word = ext.runtime_library_dirs
                elif word == "-Xlinker":
                    append_next_word = ext.extra_link_args
                elif word == "-Xcompiler":
                    append_next_word = ext.extra_compile_args
                elif switch == "-u":
                    ext.extra_link_args.append(word)
                    if not value:
                        append_next_word = ext.extra_link_args
                elif suffix in (".a", ".so", ".sl", ".o", ".dylib"):
                    # NB. a really faithful emulation of makesetup would
                    # append a .o file to extra_objects only if it
                    # had a slash in it; otherwise, it would s/.o/.c/
                    # and append it to sources.  Hmmmm.
                    ext.extra_objects.append(word)
                else:
                    file.warn(f"unrecognized argument '{word}'")

            extensions.append(ext)
    finally:
        file.close()

    return extensions



from typing import (
    Dict,
    List,
    Optional,
    Union,
)

from ._RuptureEventTypeDto import RuptureEventTypeDto


class RuptureEventDto(object):
    def __init__(
        self,                       # type: RuptureEventDto
        id,                         # type: int
        type,                       # type: RuptureEventTypeDto
        user_id,                    # type: Optional[int]
        enabled,                    # type: List[int]
        main_disabled_part_codes,   # type: List[int]
        affected_items,             # type: Dict[int, List[int]]
    ):
        # type: (...) -> None

        self.id = id
        self.type = type
        self.user_id = user_id
        self.enabled = enabled
        self.main_disabled_part_codes = main_disabled_part_codes
        self.affected_items = affected_items

    def to_dict(
        self,   # type: RuptureEventDto
    ):
        # type: (...) -> Dict[str, Union[str, int, List[int], Dict[int, List[int]]]]

        return {
            "id": self.id,
            "type": self.type.name,
            "userId": self.user_id,
            "enabled": self.enabled,
            "affectedItems": self.affected_items,
            "mainDisabledPartCodes": self.main_disabled_part_codes,
        }

    @staticmethod
    def build_from_dict(
        data_dict,  # type: Dict
    ):
        # type: (...) -> RuptureEventDto

        return RuptureEventDto(
            id=data_dict["id"],
            type=RuptureEventTypeDto(data_dict["type"]),
            user_id=data_dict.get("userId"),
            enabled=data_dict["enabled"],
            main_disabled_part_codes=data_dict["mainDisabledPartCodes"],
            affected_items={int(key): value for key, value in data_dict["affectedItems"].items()},
        )



import cfgtools

from typing import (
    Any,
    Dict,
    Optional,
    Union,
)


class Modifier(object):
    def __init__(
        self,
        key,                # type: Any[Union[cfgtools.Key, Dict]]
        from_tuple=False,   # type: Optional[bool]
    ):
        # type: (...) -> Modifier

        self.name = ''
        self.description = 0

        self.parse_modifier(
            modifier_key=key,
            from_tuple=from_tuple,
        )

    def parse_modifier(
        self,
        modifier_key,   # type: Any[Union[cfgtools.Key, Dict]]
        from_tuple,     # type: bool
    ):
        # type: (...) -> None

        self.name = modifier_key[0] if from_tuple else modifier_key.name
        self.description = modifier_key[1] if from_tuple else modifier_key.values[0]



from typing import Optional

from sysactions import (
    action,
    get_model,
    get_posot,
    show_messagebox,
    StopAction,
)
from systools import sys_log_exception
from application.interactor import OrderInteractor

from order_api.dto.exception import ProductInRupture
from actions.do_complete_option.drive_thru import validate_can_change_order
from utilfunctions import is_table_service


@action
def doModifier(
    pos_id,         # type: str
    item_id,        # type: str
    level,          # type: str
    part_code,      # type: str
    qty,            # type: str
    line_number,    # type: str
    mod_type="",    # type: Optional[str]
):
    # type: (...) -> None

    from posactions import is_changing_order

    try:
        model = get_model(posid=pos_id)

        if is_table_service(model=model) and is_changing_order(model=model):
            show_messagebox(pos_id, message="$CANNOT_ADD_ITEMS_IN_AN_OPENED_ORDER", icon="error")
            raise StopAction()

        abs_item_id = "{}.{}".format(
            item_id,
            part_code,
        )
        _validate_if_can_modify(
            pos_id=pos_id,
            item_id=abs_item_id,
            line_number=line_number,
        )

        _do_modify(
            pos_id=pos_id,
            model=model,
            item_id=item_id,
            level=level,
            part_code=part_code,
            qty=qty,
            line_number=line_number,
            mod_type=mod_type,
        )

    except ProductInRupture as ex:
        sys_log_exception("[change_modifier] Product {} in Rupture".format(ex.error_description))
        show_messagebox(
            posid=pos_id,
            message="$RUPTURED_PRODUCT|{}".format(ex.error_description),
        )


def _do_modify(
    pos_id,         # type: str
    model,          # type: str
    item_id,        # type: str
    level,          # type: str
    part_code,      # type: str
    qty,            # type: str
    line_number,    # type: str
    mod_type,       # type: str
):
    # type: (...) -> None

    from posactions import (
        get_next_level,
        change_modifier,
        get_order_sale_line,
        comment_map,
        remove_custom_comment,
        add_custom_comment,
    )

    pos_ot = get_posot(model=model)
    next_level = get_next_level(
        model=model,
        item_id=item_id,
        line_number=line_number,
        level=level,
    )
    removing = int(qty) == 0 or not mod_type
    order = change_modifier(
        pos_id=pos_id,
        line_number=line_number,
        item_id=item_id,
        level=next_level,
        part_code=part_code,
        new_qty=qty,
    )

    curr_line = get_order_sale_line(
        order=order,
        item_id=item_id,
        part_code=part_code,
        line_number=line_number,
    )
    line_exists = curr_line and len(curr_line) > 0

    comment_keys = comment_map.keys()
    if mod_type in comment_keys or removing:
        remove_custom_comment(
            pos_ot=pos_ot,
            curr_line=curr_line,
        )

        if not removing:
            if not line_exists:
                change_modifier(
                    pos_id=pos_id,
                    line_number=line_number,
                    item_id=item_id,
                    level=next_level,
                    part_code=part_code,
                    new_qty="0",
                )
                change_modifier(
                    pos_id=pos_id,
                    line_number=line_number,
                    item_id=item_id,
                    level=next_level,
                    part_code=part_code,
                    new_qty=qty,
                )

            add_custom_comment(
                pos_ot=pos_ot,
                item_id=item_id,
                part_code=part_code,
                line_number=line_number,
                level=str(next_level),
                comment=comment_map[mod_type],
            )


def _validate_if_can_modify(
    pos_id,         # type: str
    item_id,        # type: str
    line_number,    # type: str
):
    # type: (...) -> None

    order_interactor = OrderInteractor(pos_id=pos_id)
    order_picture = order_interactor.get_order_picture(pos_id=pos_id)
    if order_picture is None:
        return

    validate_can_change_order(
        pos_id=pos_id,
        order_picture=order_picture,
        line_to_replace="",
        context=item_id,
        line_number=line_number,
    )



# -*- coding: utf-8 -*-

import json

from api.product import get_product_service
from sysactions import (
    StopAction,
    action,
    show_messagebox,
)

from model.customexception import UpdateRuptureException
from .rupture_common import get_rupture_user_id_with_validation
from .. import logger


@action
def save_rupture(
    pos_id,             # type: str
    enabled_parts,      # type: str
    disabled_parts,     # type: str
):
    # type: (...) -> None

    try:
        enabled_codes_json = json.loads(enabled_parts)
        disabled_codes_json = json.loads(disabled_parts)

        user_id = int(get_rupture_user_id_with_validation(pos_id=pos_id))
        product_service = get_product_service()
        product_service.update_items(
            user_id=user_id,
            enabled=list(map(lambda item: item["product_code"], enabled_codes_json)),
            disabled=list(map(lambda item: item["product_code"], disabled_codes_json)),
        )

        show_messagebox(
            posid=pos_id,
            message="$RUPTURE_UPDATE_SUCCESS",
        )
    except StopAction:
        pass
    except (Exception, UpdateRuptureException) as e:
        logger.exception('[save_rupture] Exception {}'.format(e))
        show_messagebox(pos_id, "$ERROR_GETTING_DATA", buttons="$EXIT")



import logging
import time

from bustoken import TK_ELECTRONIC_PAYMENT_RELEASE_WAIT_CONFIRMATION
from msgbus import FM_PARAM, TK_SYS_ACK
from sysactions import mbcontext

from actions.eft import get_electronic_payment_component_name

logger = logging.getLogger("EftActions")


def release_wait_confirmation(
    pos_id,     # type: str
    status,     # type: str
    data="",    # type: str
):
    # type: (...) -> None

    sitef_comp_name = get_electronic_payment_component_name(pos_id=pos_id)
    data = "{};{};{}".format(str(pos_id), status, data)
    for _ in range(1, 4):
        try:
            ret = mbcontext.MB_EasySendMessage(
                dest_name=sitef_comp_name,
                token=TK_ELECTRONIC_PAYMENT_RELEASE_WAIT_CONFIRMATION,
                format=FM_PARAM,
                data=data,
                timeout=600000000,
            )
            if ret.token == TK_SYS_ACK:
                logger.info("Success confirming finished eft transaction. {}".format(pos_id))
                return

            message = "ACK not received when confirming finish. {} - {} - {}"
            logger.info(message.format(pos_id, ret.token, ret.data))
            time.sleep(1)

        except (Exception,):
            logger.info("Error confirming finish")



import os
import shutil
import zipfile
import time
import sqlite3

from threading import Lock
from systools import sys_parsefilepath
from typing import List

from application.model import UpdateData, UpdateType, UpdateStep, UpdateStatus
from application.model.configuration import Configurations
from application.model.customexception import (
    DisabledFunctionality,
    DataBasesNotFound,
    ErrorApplyingUpdate,
    ErrorCheckingGenesisDir,
    ErrorCheckingPackageIntegrity,
    ErrorInsertingNewUpdateVersion,
    ErrorIntegrityDatabase,
    ErrorRetrievingPendingUpdate,
    NoCatalogDirectoryFoundException,
    NoUpdateFound,
    NotInsideUpdateWindow,
)
from application.repository.api import CatalogUpdaterAPIRepository
from application.repository.pos import CatalogUpdaterPOSRepository
from application.service import UpdaterService, StaticConfigurationService
from application.util import LoggerUtil, settings

logger = LoggerUtil.get_logger_name(__name__)
GENESIS_PATH = os.path.join("..", "..", "genesis")
HT_DOCS_POS = "htdocs/pos"


class CatalogUpdaterInteractor(object):

    def __init__(self, configs, working_on_update, api_repository, pos_repository, updater_service, enabled):
        # type: (Configurations, Lock, CatalogUpdaterAPIRepository, CatalogUpdaterPOSRepository, UpdaterService, bool) -> None # noqa

        self.working_on_update = working_on_update
        self.update_type = UpdateType.catalog
        self.update_name = self.update_type.name
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.updater_service = updater_service
        self.enabled = enabled

        data_path = sys_parsefilepath("{HVDATADIR}")
        work_dir = sys_parsefilepath(configs.backup_directory)

        self.databases_path = os.path.join(data_path, "databases")
        self.bundles_path = os.path.join(data_path, "bundles")
        self.htdocs_pos_path = os.path.join(data_path, HT_DOCS_POS)
        self.genesis_databases_path = os.path.join(data_path, GENESIS_PATH, "data/server/databases")
        self.genesis_campaign_databases_path = os.path.join(data_path, GENESIS_PATH, "data/server/databases-campaign")
        self.genesis_htdocs_pos_path = os.path.join(data_path, GENESIS_PATH, "data/server/htdocs/pos")
        self.catalog_json = os.path.join(self.genesis_campaign_databases_path, "catalog.json")

        default_work_path = os.path.join(work_dir, self.update_name)
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.backups_path = os.path.join(default_work_path, "backups")
        self.work_path = os.path.join(default_work_path, "work_path")

        self.updater_service.create_dir(self.backups_path, False)
        self.updater_service.create_dir(self.downloads_path, False)

    def apply_update(
        self,               # type: CatalogUpdaterInteractor
        force=False,        # type: bool
    ):
        # type: (...) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        try:
            settings.set_update_control(self.update_name, "$STARTING_CATALOG_UPDATE")
            settings.set_update_control("updating", True)
            update_data = self.pos_repository.has_catalog_pending_update_to_apply()
            if not force and not update_data:
                return

            logger.info("Waiting Updating Catalog")
            with self.working_on_update:
                logger.info("Applying update")
                settings.set_update_control(self.update_name, "$GETTING_MENU_FROM_BACKOFFICE")
                self._create_backup(update_data)
                self._apply_update(update_data)
                self._copy_upsell_part_codes_json()

                self.updater_service.update_databases_with_json_config(
                    update_name=self.update_name,
                    settings=settings,
                )

                StaticConfigurationService.update_static_databases()
                StaticConfigurationService.run_databases_script()
                logger.info("Update [{}] fully concluded with success!".format(update_data.update_id))

                logger.info("Skipping pending updates")
                self._skip_pending_updates()

                logger.info("Restarting Hypervisor to apply new Menu")
                settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")
                self.pos_repository.common_pos_repository.restart_hv(
                    event_type="POS_UPDATER_CATALOG",
                )
                time.sleep(5)
        except NoUpdateFound:
            logger.info("No update founded to apply")
            raise
        except NotInsideUpdateWindow:
            logger.info("Sleeping until enter the update window")
            raise
        except ErrorRetrievingPendingUpdate:
            logger.exception("Error retrieving pending update")
            raise
        except ErrorCheckingGenesisDir:
            logger.error("Cannot find the server genesis databases directory. Please, verify and try again")
            raise
        except ErrorInsertingNewUpdateVersion:
            logger.exception("Error inserting new update")
            raise
        except ErrorIntegrityDatabase:
            settings.set_update_control(self.update_name, "$ERROR_INTEGRITY_DATABASE")
            raise
        except (Exception,):
            logger.exception("Not expected exception on catalog update")
            settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")
            raise
        finally:
            settings.set_update_control("updating", False)

    def _apply_update(self, update_data):
        # type: (UpdateData) -> None
        try:
            if update_data.applied_date:
                return

            logger.info("Applying: get menus from catalog...")
            database_version = self.pos_repository.get_last_downloaded_version(UpdateType.catalog)
            database_version_name = database_version["updateName"]
            database_id = self.updater_service.get_database_id(database_version_name)

            self._copy_all_databases(database_id)
            self.pos_repository.update_step(update_data.update_id, UpdateStep.apply, UpdateType.catalog)
            logger.info("Menus successfully copied!")
        except ErrorIntegrityDatabase as error:
            self._restore_backup(update_data)
            raise error
        except Exception as ex:
            logger.exception("Error applying get menus from catalog")
            self._restore_backup(update_data)
            raise ErrorApplyingUpdate(ex)

    def _restore_backup(self, update_data):
        # type: (UpdateData) -> None

        logger.info("Restoring backup...")

        self.pos_repository.update_status(update_data.update_id, UpdateStatus.error)
        package_name = self.updater_service.get_package_name(update_data)
        backup_package = os.path.join(self.backups_path, package_name)

        self._perform_update(backup_package)

        logger.info("Backup successfully restored!")

    def _skip_pending_updates(self, type_id=None):
        # type: (UpdateType) -> None

        logger.info("Skipping old pending updates...")
        update_value = self.update_type.value if type_id is None else type_id
        self.pos_repository.skip_pending_updates(str(update_value))

    def _create_work_dir(self):
        # type: () -> None

        if os.path.exists(self.work_path):
            self._delete_work_dir()

        os.mkdir(self.work_path)

    def _delete_work_dir(self):
        # type: () -> None

        shutil.rmtree(self.work_path)

    def _get_files_to_backup(self):
        # type: () -> List[str]

        self.updater_service.create_dir(self.genesis_databases_path, False)
        self.updater_service.create_dir(self.genesis_htdocs_pos_path, False)

        databases_list = os.listdir(os.path.join(self.genesis_databases_path))
        htdocs_list = os.listdir(os.path.join(self.genesis_htdocs_pos_path))

        return databases_list + htdocs_list

    def _copy_files_to_backup(self, files_to_backup):
        # type: (List[str]) -> None

        logger.info("Coping backup files to work dir...")
        for file_to_backup in files_to_backup:
            if file_to_backup.endswith(".db") or file_to_backup.endswith(".json"):
                file_path = os.path.join(self.genesis_databases_path, file_to_backup)
                databases_pos_path = os.path.join(self.work_path, "databases")
                self.updater_service.create_dir(databases_pos_path, False)
                shutil.copy(file_path, databases_pos_path)
            elif file_to_backup.endswith(".xml"):
                file_path = os.path.join(self.htdocs_pos_path, file_to_backup)
                if os.path.exists(file_path):
                    work_htdocs_pos_path = os.path.join(self.work_path, HT_DOCS_POS)
                    self.updater_service.create_dir(work_htdocs_pos_path, False)
                    shutil.copy(file_path, work_htdocs_pos_path)
            else:
                logger.error("Unknown file extension on backup: {}".format(file_to_backup))

        logger.info("Backup files successfully copied!")

    def _write_backup_zip(self, backup_package, files_to_backup):
        # type: (str, List[str]) -> None

        logger.info("Zipping backup files...")
        with zipfile.ZipFile(backup_package, "w") as zip_object:
            for file_to_backup in files_to_backup:
                if file_to_backup.endswith(".db") or file_to_backup.endswith(".json"):
                    path_to_store = os.path.join("databases", file_to_backup)
                elif file_to_backup.endswith(".xml"):
                    path_to_store = os.path.join(HT_DOCS_POS, file_to_backup)
                else:
                    logger.error("Unknown file extension on create zip: {}".format(file_to_backup))
                    continue

                file_path = os.path.join(self.work_path, path_to_store)
                if os.path.exists(file_path):
                    zip_object.write(file_path, path_to_store, zipfile.ZIP_DEFLATED)
        logger.info("Backup files successfully zipped!")

    def _perform_update(self, package):
        # type: (str) -> None

        self._create_work_dir()
        self._extract_package_on_work_dir(package)
        self._copy_working_files_to_genesis()
        self._delete_work_dir()

    def _extract_package_on_work_dir(self, package):
        # type: (str) -> None

        logger.info("Extracting package on work dir...")
        with zipfile.ZipFile(package, "r") as zip_object:
            zip_object.extractall(self.work_path)
        logger.info("Package successfully extracted!")

    def _copy_working_files_to_genesis(self):
        # type: () -> None

        logger.info("Coping files to genesis...")
        files_to_apply = os.listdir(self.work_path)
        for file_to_apply in files_to_apply:
            file_path = os.path.join(self.work_path, file_to_apply)
            if file_path.endswith(".db"):
                shutil.copy(file_path, self.genesis_databases_path)
            elif file_path.endswith(".xml"):
                shutil.copy(file_path, self.genesis_htdocs_pos_path)
            else:
                logger.error("Unknown file extension apply update: {}".format(file_to_apply))

        logger.info("Files copied with success!")

    def _create_backup(self, update_data):
        # type: (UpdateData) -> None

        if update_data.backup_date:
            return

        logger.info("Creating backup...")
        settings.set_update_control(self.update_name, "$CREATING_MENU_BACKUP")

        package_name = self.updater_service.get_package_name(update_data)
        backup_package = os.path.join(self.backups_path, package_name)

        self._create_work_dir()
        files_to_backup = self._get_files_to_backup()
        self._copy_files_to_backup(files_to_backup)
        self._write_backup_zip(backup_package, files_to_backup)
        self.updater_service.test_zip_file(backup_package)
        self._delete_work_dir()

        self.pos_repository.update_step(update_data.update_id, UpdateStep.backup, UpdateType.catalog)

        logger.info("Backup successfully created!")

    def _fix_catalog_directory(self):
        try:
            json_to_download = self.updater_service.retrieve_pending_update()
            self.updater_service.download_databases_campaign(json_to_download, False)
        except ErrorCheckingPackageIntegrity as error:
            raise error
        except (Exception,):
            raise NoCatalogDirectoryFoundException()

    @staticmethod
    def _check_database_integrity(
        db_path,    # type: str
    ):
        # type: (...) -> None

        con = sqlite3.connect(db_path)
        cur = con.cursor()
        try:
            status = cur.execute("PRAGMA integrity_check").fetchone()
            if status[0] != 'ok':
                logger.error("Error of integrity of database: {}. Status: {}".format(db_path, status[0]))
                raise ErrorIntegrityDatabase()
        except Exception as ex:
            logger.error("Error of integrity of database: {}. Exception: {}".format(db_path, ex.message))
            raise ErrorIntegrityDatabase()

    def _copy_all_databases(self, database_id):
        # type: (int) -> None

        logger.info("Applying: start copy databases campaign, database to apply: {}...".format(database_id))

        source_path = "{}/{}".format(self.genesis_campaign_databases_path, database_id)
        if not os.path.isdir(source_path):
            self._fix_catalog_directory()

        src_files = os.listdir(source_path)
        if len(src_files) == 0:
            raise DataBasesNotFound()

        for file_name in src_files:
            full_file_name = os.path.join(source_path, file_name)
            if os.path.isfile(full_file_name):
                shutil.copy(full_file_name, self.genesis_databases_path)

                if file_name.endswith('.db'):
                    db_path = os.path.join(self.genesis_databases_path, file_name)
                    self._check_database_integrity(db_path=db_path)

        logger.info("Applying: copy databases campaign finish...")

    def _copy_upsell_part_codes_json(
        self
    ):
        # type: (...) -> None

        try:
            origin_upsell_part_codes_json = os.path.join(self.genesis_databases_path, "upSellPartCodes.json")
            if not os.path.exists(origin_upsell_part_codes_json):
                logger.info("The file {} was not found to update".format(origin_upsell_part_codes_json))
                return

            internal_order_dir_exists = os.path.exists(
                os.path.join(
                    self.bundles_path,
                    "internalorder",
                )
            )
            if internal_order_dir_exists:
                destination_upsell_part_codes_json = os.path.join(
                    self.bundles_path,
                    "internalorder",
                    "upSellPartCodes.json"
                )
                self.updater_service.copy_file(
                    original_file_path=origin_upsell_part_codes_json,
                    destination_file_path=destination_upsell_part_codes_json,
                )
        except (Exception,):
            logger.exception("Error coping upSellPartCodes.json")



import ast
import base64
import json
from datetime import datetime
from typing import (
    List,
    Optional,
    Dict,
    Union,
)
from uuid import UUID

import iso8601
from event_manager_api.shared.error import (
    CANNOT_PROCESS_SYNC_EVENT,
    ASYNC_TARGET_MUST_HAVE_DATA,
    INVALID_UTF8_DATA,
    ASYNC_TARGET_MUST_HAVE_JSON_DATA,
    ASYNC_TARGET_MUST_DATA_MUST_CONTAIN_AN_ID,
    CANNOT_PROCESS_TARGET_AND_ASYNC_TARGET,
)
from event_manager_api.exception import (
    CannotProcessSyncEvent,
    AsyncTargetForEventWithNoData,
    AsyncTargetForEventWithNoUtf8Data,
    AsyncTargetForEventWithNoJsonData,
    AsyncTargetForEventWithDataWithoutId,
    CannotProcessATargetAndAsyncTarget,
)
from event_manager_api.shared.token import (
    TK_EVENT_MANAGER_PUBLISH,
    TK_EVENT_MANAGER_GET_PENDING,
    TK_EVENT_MANAGER_GET_ABANDONED,
    TK_EVENT_MANAGER_GET_DETAILS,
    TK_EVENT_MANAGER_DEFINITION_NOT_FOUND,
)
from messagebus import (
    MessageBus,
    Event,
    Message,
    DataType,
    DefaultToken,
    BadRequestMessage,
)
from event_manager_api.shared._AsyncTarget import AsyncTarget
from event_manager_api.shared._EventPublishDetailDto import EventPublishDetailDto
from event_manager_api.shared._PublicationAction import PublicationAction
from event_manager_api.shared._PublicationHistoryDto import PublicationHistoryDto


class EventManager(object):
    def __init__(
        self,               # type: EventManager
        message_bus,        # type: MessageBus
    ):
        self.message_bus = message_bus

        self.error_code_exception_map = {
            CANNOT_PROCESS_SYNC_EVENT[0]: CannotProcessSyncEvent,
            ASYNC_TARGET_MUST_HAVE_DATA[0]: AsyncTargetForEventWithNoData,
            INVALID_UTF8_DATA[0]: AsyncTargetForEventWithNoUtf8Data,
            ASYNC_TARGET_MUST_HAVE_JSON_DATA[0]: AsyncTargetForEventWithNoJsonData,
            ASYNC_TARGET_MUST_DATA_MUST_CONTAIN_AN_ID[0]: AsyncTargetForEventWithDataWithoutId,
            CANNOT_PROCESS_TARGET_AND_ASYNC_TARGET[0]: CannotProcessATargetAndAsyncTarget,
        }

    def publish(
        self,                    # type: EventManager
        event,                   # type: Event,
        target=None,             # type: Optional[str]
        async_target=None,       # type: Optional[AsyncTarget]
        external_code=None,      # type: Optional[str]
        fifo=False,              # type: bool
    ):
        # type: (...) -> EventPublishDetailDto

        request = self._build_request(
            event=event,
            target=target,
            async_target=async_target,
            external_code=external_code,
            fifo=fifo,
        )

        reply = self.message_bus.send_message(
            component="CustomEventManager",
            message=Message(
                token=TK_EVENT_MANAGER_PUBLISH,
                data_type=DataType.param,
                data=json.dumps(request).encode("utf-8"),
            )
        )

        if reply.token == DefaultToken.TK_SYS_ACK:
            return self._build_event_detail_dto(reply)

        raise self._build_exception_from_reply(reply)

    def publish_list(
        self,                    # type: EventManager
        event,                   # type: Event,
        targets=None,            # type: List[str]
        async_target=None,       # type: Optional[AsyncTarget]
        external_code=None,      # type: Optional[str]
        fifo=False,              # type: bool
    ):
        # type: (...) -> List[Union[EventPublishDetailDto, Exception]]

        detail_return = []
        for target in targets:
            ret = self.publish(
                event=event,
                target=target,
                async_target=async_target,
                external_code=external_code,
                fifo=fifo,
            )

            detail_return.append(ret)

        return detail_return

    def get_pending(self):
        # type: (...) -> List[EventPublishDetailDto]

        reply = self.message_bus.send_message(
            component="CustomEventManager",
            message=Message(
                token=TK_EVENT_MANAGER_GET_PENDING,
                data_type=DataType.string,
                data=b"",
            )
        )
        if reply.token != DefaultToken.TK_SYS_ACK:
            raise self._build_internal_error(message=reply)

        return self._build_event_detail_dto_list(reply)

    def get_abandoned(
        self,       # type: EventManager
    ):
        # type: (...) -> List[EventPublishDetailDto]

        reply = self.message_bus.send_message(
            component="CustomEventManager",
            message=Message(
                token=TK_EVENT_MANAGER_GET_ABANDONED,
                data_type=DataType.string,
                data=b"",
            )
        )
        if reply.token != DefaultToken.TK_SYS_ACK:
            raise self._build_internal_error(message=reply)

        return self._build_event_detail_dto_list(reply)

    def get_details(
        self,               # type: EventManager
        publish_detail_id,  # type: UUID
    ):
        # type: (...) -> Optional[EventPublishDetailDto]

        reply = self.message_bus.send_message(
            component="CustomEventManager",
            message=Message(
                token=TK_EVENT_MANAGER_GET_DETAILS,
                data_type=DataType.string,
                data=str(publish_detail_id).encode("utf-8"),
            )
        )
        if reply.token == TK_EVENT_MANAGER_DEFINITION_NOT_FOUND:
            return None

        if reply.token != DefaultToken.TK_SYS_ACK:
            raise self._build_internal_error(message=reply)

        return self._build_event_detail_dto(reply=reply)

    def _build_exception_from_reply(
        self,       # type: EventManager,
        reply,      # type: Message
    ):
        # type: (...) -> Exception

        if reply.token == DefaultToken.TK_SYS_BAD_REQUEST:
            bad_request = BadRequestMessage.build_from_message(message=reply)
            if bad_request.error_code in self.error_code_exception_map:
                return self.error_code_exception_map[bad_request.error_code]()

        return self._build_internal_error(message=reply)

    @staticmethod
    def _build_internal_error(
        message,        # type: Message
    ):
        # type: (...) -> RuntimeError

        try:
            data_text = message.data.decode("utf-8")
        except UnicodeDecodeError:
            data_text = message.data.hex()
        return RuntimeError("Invalid response: {} - {}".format(message.token, data_text))

    def _build_request(
        self,                   # type: EventManager
        event,                  # type: Event
        target=None,            # type: Optional[str]
        async_target=None,      # type: Optional[AsyncTarget]
        external_code=None,     # type: Optional[str]
        fifo=False,             # type: bool
    ):
        # type: (...) -> Dict

        request = {
            "event": self._convert_event(
                event=event,
            )
        }
        if target is not None:
            request["target"] = target
        if async_target is not None:
            request["asyncTarget"] = {
                "replySubject": async_target.reply_subject,
            }
            if async_target.id_field != "id":
                request["asyncTarget"]["idField"] = async_target.id_field
        if external_code is not None:
            request["externalCode"] = external_code

        request["fifo"] = fifo

        return request

    def _build_event_detail_dto_list(
        self,           # type: EventManager
        message,        # type: Message
    ):
        # type: (...) -> List[EventPublishDetailDto]

        definitions_dict = json.loads(message.data.decode("utf-8"))
        ret = []
        for definition in definitions_dict:
            ret.append(self._build_event_detail_from_dict(definition))
        return ret

    def _build_event_detail_dto(
        self,       # type: EventManager
        reply,      # type: Message
    ):
        # type: (...) -> EventPublishDetailDto

        definition_dict = json.loads(reply.data.decode("utf-8"))
        return self._build_event_detail_from_dict(event_detail=definition_dict)

    def _build_event_detail_from_dict(
        self,           # type: EventManager
        event_detail,   # type: Dict
    ):
        # type: (...) -> EventPublishDetailDto

        return EventPublishDetailDto(
            id=event_detail["id"],
            external_code=event_detail.get("externalCode"),
            event=(self._build_event_from_dict(event_dict=event_detail["event"])),
            acked_at=self._build_optional_datetime(event_detail.get("ackedAt")),
            next_retry_at=self._build_optional_datetime(event_detail.get("nextRetryAt")),
            abandoned_at=self._build_optional_datetime(event_detail.get("abandonedAt")),
            publication_history=self._build_publication_history_from_list(event_detail.get("publicationHistory", [])),
            fifo=event_detail["fifo"],
            target=event_detail["target"],
        )

    @staticmethod
    def _build_event_from_dict(
        event_dict,     # type: Dict
    ):
        # type: (...) -> Event

        data = None
        if "data" in event_dict:
            data = base64.b64decode(event_dict["data"].encode("utf-8"))
        return Event(
            subject=event_dict["subject"],
            event_type=event_dict["eventType"] if "eventType" in event_dict else None,
            data=data,
            headers=ast.literal_eval(event_dict["headers"]),
        )

    @staticmethod
    def _build_publication_history_from_list(
        publication_history,        # type: List[Dict]
    ):
        # type: (...) -> List[PublicationHistoryDto]

        ret = []
        for history_dict in publication_history:
            ret.append(
                PublicationHistoryDto(
                    created_at=iso8601.parse_date(history_dict["createdAt"]),
                    action=PublicationAction(history_dict["action"])
                )
            )

        return ret

    @staticmethod
    def _convert_event(
        event,      # type: Event
    ):
        # type: (...) -> Dict

        event_json = {
            "subject": event.subject,
            "type": event.event_type,
            "headers": str(event.headers),
        }

        data = event.data
        if event.data is not None:
            if not isinstance(data, (bytes, bytearray)):
                try:
                    data = data.encode()
                except (AttributeError, UnicodeDecodeError):
                    pass

            event_json["data"] = base64.b64encode(data).decode("utf-8")

        if event.synchronous is True:
            event_json["synchronous"] = True

        return event_json

    @staticmethod
    def _build_optional_datetime(
        value,      # type: str
    ):
        # type: (...) -> Optional[datetime]

        if value is None:
            return None
        return iso8601.parse_date(value)



# vim:ts=4 sw=4 expandtab softtabstop=4
import argparse
import io
import locale
import os
import sys

from unidecode import unidecode

def fatal(msg):
    sys.stderr.write(msg + "\n")
    sys.exit(1)

def main():
    default_encoding = locale.getpreferredencoding()

    parser = argparse.ArgumentParser(
            description="Transliterate Unicode text into ASCII. FILE is path to file to transliterate. "
            "Standard input is used if FILE is omitted and -c is not specified.")
    parser.add_argument('-e', '--encoding', metavar='ENCODING', default=default_encoding,
            help='Specify an encoding (default is %s)' % (default_encoding,))
    parser.add_argument('-c', metavar='TEXT', dest='text',
            help='Transliterate TEXT instead of FILE')
    parser.add_argument('path', nargs='?', metavar='FILE')

    args = parser.parse_args()

    encoding = args.encoding

    if args.path:
        if args.text:
            fatal("Can't use both FILE and -c option")
        else:
            stream = open(args.path, 'rb')
    elif args.text:
        text = os.fsencode(args.text)
        # add a newline to the string if it comes from the
        # command line so that the result is printed nicely
        # on the console.
        stream = io.BytesIO(text + b'\n')
    else:
        stream = sys.stdin.buffer

    for line_nr, line in enumerate(stream):
        try:
            line = line.decode(encoding)
        except UnicodeDecodeError as e:
            fatal('Unable to decode input line %s: %s, start: %d, end: %d' % (line_nr, e.reason, e.start, e.end))

        sys.stdout.write(unidecode(line))
    stream.close()



from threading import Lock
from typing import List

from msgbus import MBEasyContext
from mwhelper import BaseRepository
from persistence import Connection

from dailygoals_services.model import SaleChallenge
from dailygoals_services.model import GetSalesChallengeException, InsertSalesChallengeException, \
    DeleteSalesChallengeException, UpdateSalesChallengeException
from dailygoals_services.model.repository import DailyGoalsRepository as DailyGoalsServicesRepository


class DailyGoalsRepository(DailyGoalsServicesRepository, BaseRepository):

    def __init__(self, mb_context, persistence_name, database_lock):
        # type: (MBEasyContext, str, Lock) -> None

        super(DailyGoalsRepository, self).__init__(mb_context)

        self.persistence_name = persistence_name
        self.database_lock = database_lock

    def get_sales_challenge(self, limit=None):
        # type: (int) -> List[SaleChallenge]

        limit = limit if limit else 5
        try:
            def inner_func(conn):
                # type: (Connection) -> List[SaleChallenge]

                query = """
                            SELECT Id, Name, StartDate, EndDate, Products, CreatedBy
                            FROM SalesChallenge
                            ORDER BY StartDate DESC, EndDate DESC
                            LIMIT {}
                        """
                query_format = query.format(limit)
                cursor = conn.select(query_format)
                sales_challenge = []

                for row in cursor:
                    products = row.get_entry("Products") or "[]"
                    challenge = SaleChallenge(
                        name=row.get_entry("Name").decode("utf-8"),
                        start_date=row.get_entry("StartDate").decode(),
                        end_date=row.get_entry("EndDate").decode(),
                        products=products,
                        created_by=row.get_entry("CreatedBy"),
                        challenge_id=int(row.get_entry("Id")),
                    )
                    sales_challenge.append(challenge)
                return sales_challenge

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except (Exception,):
            raise GetSalesChallengeException()

    def insert_sales_challenge(self, challenge):
        # type: (SaleChallenge) -> int

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            INSERT INTO SalesChallenge(
                                Name, StartDate, EndDate, Products, CreatedBy
                            )
                            VALUES ('{}', '{}', '{}', '{}', '{}')
                        """
                query_format = query.format(
                    challenge.name.encode("utf-8"),
                    challenge.start_date.encode(),
                    challenge.end_date.encode(),
                    challenge.products,
                    challenge.created_by,
                )
                return conn.query(query_format)

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except (Exception,):
            raise InsertSalesChallengeException()

    def delete_sales_challenge(self, sale_challenge_id):
        # type: (int) -> int

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            DELETE FROM SalesChallenge
                            WHERE Id={}
                        """
                query_format = query.format(sale_challenge_id)
                return conn.query(query_format)

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except (Exception,):
            raise DeleteSalesChallengeException()

    def update_sales_challenge(self, challenge):
        # type: (SaleChallenge) -> int

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                            UPDATE SalesChallenge
                            SET Name='{}', StartDate='{}', EndDate='{}', Products='{}', CreatedBy={}
                            WHERE Id={}
                        """
                query_format = query.format(
                    challenge.name.encode("utf-8"),
                    challenge.start_date.encode(),
                    challenge.end_date.encode(),
                    challenge.products,
                    challenge.created_by,
                    challenge.challenge_id,
                )
                return conn.query(query_format)

            return self.execute_with_connection(inner_func, service_name=self.persistence_name)
        except (Exception,):
            raise UpdateSalesChallengeException()



from datetime import datetime
from typing import (
    Dict,
    List,
    Optional,
)

from xml.etree import cElementTree as eTree

from ._OrderFormatter import OrderFormatter
from ._Type import JSON
from ._util import get_tender_detail


class TableFormatter:

    def __init__(self):
        return

    def get_table_picture_dict(
        self,
        table_picture,          # type: eTree.Element
    ):
        # type: (...) -> JSON

        start_ts = table_picture.get("startTS")
        start_ts = self._get_time_stamp(start_ts)

        finish_ts = table_picture.get("finishTS")
        finish_ts = self._get_time_stamp(finish_ts)

        linked_tables = table_picture.get("linkedTables")
        linked_tables = linked_tables.split(",") if linked_tables is not None else []

        table_picture_dict = {
            "id": table_picture.get("id"),
            "businessPeriod": table_picture.get("businessPeriod"),
            "calculatedTip": float(table_picture.get("calculatedTip", 0)),
            "lastUpdateTS": table_picture.get("lastUpdateTS"),
            "name": table_picture.get("name"),
            "seats": int(table_picture.get("seats", 1)),
            "sector": table_picture.get("sector"),
            "serviceId": int(table_picture.get("serviceId", 0)) or None,
            "serviceType": table_picture.get("type"),
            "serviceTypeDesc": table_picture.get("typeDescr"),
            "serviceSeats": int(table_picture.get("serviceSeats", 1)),
            "startTS": start_ts,
            "finishTS": finish_ts,
            "status": table_picture.get("status"),
            "statusDesc": table_picture.get("statusDescr"),
            "sliceSourceTableId": table_picture.get("sliceSourceTableId", ""),
            "sliceParentTableId": table_picture.get("sliceParentTableId", ""),
            "sliceSourceServiceId": self._get_source_service_id(table_picture),
            "subSector": table_picture.get("subSector"),
            "userId": int(table_picture.get("userId", 0)) or None,
            "linkedTables": linked_tables,
            "totalAmount": float(table_picture.get("totalAmount", 0)),
            "discountAmount": float(table_picture.get("discountAmount", 0)),
            "tipAmount": float(table_picture.get("tipAmount")) if table_picture.get("tipAmount") else 0,
            "tipRate": float(table_picture.get("tipRate")) if table_picture.get("tipRate") else 0,
            "totalTender": float(table_picture.get("totalTender", 0)),
            "dueAmount": float(table_picture.get("dueAmount", 0)),
            "change": float(table_picture.get("change", 0)),
            "taxTotal": float(table_picture.get("taxTotal", 0)),
            "ServiceChildren": [],
            "Properties": self._get_custom_properties(table_picture),
            "ServiceOrders": self._get_orders(table_picture, "ServiceOrders/ServiceOrder"),
            "PaymentOrders": self._get_orders(table_picture, "PaymentOrders/PaymentOrder"),
            "ServiceTenders": self._get_service_tenders(table_picture),
            "Orders": self._get_orders_info(table_picture),
        }

        current_pos_id = table_picture.get("currentPOSId")
        if current_pos_id is not None:
            table_picture_dict["currentPOSId"] = current_pos_id

        return table_picture_dict

    @staticmethod
    def _get_time_stamp(time_stamp):
        if time_stamp:
            try:
                ret_time_stamp = datetime.strptime((time_stamp + "Z"), "%Y-%m-%d %H:%M:%SZ")
            except ValueError:
                try:
                    ret_time_stamp = datetime.strptime((time_stamp + "Z"), "%Y-%m-%dT%H:%M:%SZ")
                except ValueError:
                    return None
            return datetime.strftime(ret_time_stamp, "%Y-%m-%dT%H:%M:%SZ")

        return None

    @staticmethod
    def _get_custom_properties(
        table_picture           # type: eTree.Element
    ):
        # type: (...) -> Dict

        payload_custom_properties = {}
        custom_properties = table_picture.find("Properties")
        if custom_properties is not None:
            for key, value in custom_properties.items():
                payload_custom_properties[key] = value

        return payload_custom_properties

    @staticmethod
    def _get_orders(
        table_picture,      # type: eTree.Element
        tag                 # type: str
    ):
        # type: (...) -> List

        payload_orders = []
        for order_line in table_picture.findall("{}".format(tag)):
            order_id = int(order_line.get("orderId", -1))
            order_state = order_line.get("state")
            created_at = order_line.get("createdAt")
            if created_at is None:
                created_at = datetime.utcfromtimestamp(0).isoformat()

            order_payload = {
                "orderId": order_id,
                "state": order_state,
                "createdAt": created_at + "Z"
            }
            payload_orders.append(order_payload)

        return payload_orders

    @staticmethod
    def _get_orders_info(
        table_picture,  # type: eTree.Element
    ):
        # type: (...) -> List

        table_orders = table_picture.find("Orders")
        if not table_orders:
            return []

        orders = []
        order_formatter = OrderFormatter()
        for order in table_orders.findall("Order"):
            orders.append(order_formatter.get_order_picture_dict(order_picture=order))

        return orders

    @staticmethod
    def _get_source_service_id(
        table_picture       # type: eTree.Element
    ):
        # type: (...) -> Optional[int]

        if table_picture.get("sourceServiceId") is not None:
            return int(table_picture.get("sourceServiceId"))

        return None

    @staticmethod
    def _get_service_tenders(
        table_picture       # type: eTree.Element
    ):
        # type: (...) -> List

        service_tenders = []
        for tender_line in table_picture.findall("ServiceTenders/ServiceTender"):
            tender_payload = {
                "serviceTenderId": tender_line.get("serviceTenderId"),
                "orderId": tender_line.get("orderId"),
                "orderTenderId": tender_line.get("orderTenderId"),
                "tenderAmount": float(tender_line.get("tenderAmount", 0)),
                "tenderDescr": tender_line.get("TenderDescr"),
                "tenderType": tender_line.get("tenderId"),
                "changeAmount": float(tender_line.get("change", 0)),
                "tenderDetail": get_tender_detail(tender_line=tender_line),
                "tenderTip": float(tender_line.get("tip", 0))
            }
            service_tenders.append(tender_payload)

        return service_tenders



class WeightComparer(object):
    def __init__(self):
        self.previous_weight = None

    def is_weight_equal_to_previous(self, weight):
        if self.previous_weight == weight:
            return True
        else:
            self.previous_weight = weight
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Convenience module importing everything from backports.configparser."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from backports.configparser import (
    RawConfigParser,
    ConfigParser,
    SafeConfigParser,
    SectionProxy,
    Interpolation,
    BasicInterpolation,
    ExtendedInterpolation,
    LegacyInterpolation,
    NoSectionError,
    DuplicateSectionError,
    DuplicateOptionError,
    NoOptionError,
    InterpolationError,
    InterpolationMissingOptionError,
    InterpolationSyntaxError,
    InterpolationDepthError,
    ParsingError,
    MissingSectionHeaderError,
    ConverterMapping,
    DEFAULTSECT,
    MAX_INTERPOLATION_DEPTH,
)

from backports.configparser import Error, _UNSET, _default_dict, _ChainMap  # noqa: F401

__all__ = [
    "NoSectionError",
    "DuplicateOptionError",
    "DuplicateSectionError",
    "NoOptionError",
    "InterpolationError",
    "InterpolationDepthError",
    "InterpolationMissingOptionError",
    "InterpolationSyntaxError",
    "ParsingError",
    "MissingSectionHeaderError",
    "ConfigParser",
    "SafeConfigParser",
    "RawConfigParser",
    "Interpolation",
    "BasicInterpolation",
    "ExtendedInterpolation",
    "LegacyInterpolation",
    "SectionProxy",
    "ConverterMapping",
    "DEFAULTSECT",
    "MAX_INTERPOLATION_DEPTH",
]

# NOTE: names missing from __all__ imported anyway for backwards compatibility.



# -*- coding: utf-8 -*-
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

import sysactions
from msgbus import (
    TK_SYS_ACK,
    TK_SLCTRL_OMGR_ORDERPICT,
    FM_PARAM,
)
from pos_model import TableStatus
from typing import (
    Optional,
    Union,
    Any,
)

from application.repository import AccountRepository
from actions.login.pause_user import pause_user
from actions.login.common_lock import release_user_lock
from actions.functions.do_store_order import doStoreOrder
from actions.functions.do_transfer import (
    do_transfer,
    get_justification,
    get_transfer_values,
)
from actions.config import get_close_operator_conference_type
from actions.util import get_operator_info, get_user_level
from actions.void.do_void_order import do_void_order

from model.customexception import (
    CloseOrderByOperatorCloseException,
    NotFoundUserError,
)

from utilfunctions import get_user_id_from_session
from .lock_user_actions import acquire_user_lock_close_operator
from actions.login.get_lock_message import get_lock_message
from .. import logger, mb_context
from .utils import (
    close_user,
)
from actions.functions.dto import ConferenceTypes
from actions.table.store_table_service import store_table_service
from actions.order.order_was_already_saved import order_was_already_saved

USECS_PER_SEC = 1000000


def close_quick_service_order_from_another_pos(
    pos_id,                                 # type: str
    order,                                  # type: eTree.Element
    user_id,                                # type: str
    pos_to_close_user,                      # type: str
    must_store_quick_service_order=False,   # type: Optional[bool]
):
    # type: (...) -> None

    if order.get("state") not in ("IN_PROGRESS", "TOTALED"):
        return

    user_id_from_session = get_user_id_from_session(session_id=order.get("sessionId"))
    if user_id_from_session != user_id:
        return

    if must_store_quick_service_order and order_was_already_saved(order=order):
        doStoreOrder(
            pos_id=pos_to_close_user,
            pos_id_to_show_message=pos_id,
            show_popup=False,
        )
        return

    order_voided = _voided_quick_service_order_after_confirmation(
        pos_id=pos_id,
        pos_to_close_user=pos_to_close_user,
    )
    if not order_voided:
        raise CloseOrderByOperatorCloseException(
            service_type="Quick Service",
            pos_id=pos_id,
        )


def close_table_service_from_another_pos(
    pos_id,                             # type: str
    order,                              # type: eTree.Element
    user_id,                            # type: str
    pos_to_close_user,                  # type: str
):
    # type: (...) -> None

    _void_in_progress_table_order(
        order=order,
        pos_id=pos_id,
        pos_to_close_user=pos_to_close_user,
        user_id=user_id,
    )

    store_table_service(
        pos_id=pos_to_close_user,
    )


def _void_in_progress_table_order(
    order,                              # type: eTree.Element
    pos_id,                             # type: str
    pos_to_close_user,                  # type: str
    user_id,                            # type: str
):
    # type: (...) -> None

    if order.get("state") != "IN_PROGRESS":
        return

    user_id_from_session = get_user_id_from_session(session_id=order.get("sessionId"))
    if user_id_from_session != user_id:
        return

    if not order_was_already_saved(order=order):
        table_id = order.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("TABLE_ID"))
        order_voided = _voided_table_order_after_confirmation(
            pos_id=pos_id,
            pos_to_close_user=pos_to_close_user,
            order_id=order.get("orderId"),
            table_id=table_id.get("value"),
        )
        if not order_voided:
            raise CloseOrderByOperatorCloseException(
                service_type="Table Service",
                pos_id=pos_id,
            )


def _store_order_in_table_to_close_user(
    pos_id,             # type: str
    pos_to_close_user,  # type: str
    table,              # type: Any
    table_service,      # type: Any
    user_id,            # type: str
):
    # type: (...) -> bool

    if not user_id or table.status not in (TableStatus.TOTALED, TableStatus.IN_PROGRESS):
        return False

    table_picture = eTree.XML(table_service().get_table_picture(
        pos_id=pos_to_close_user,
        table_id=table.id,
    ))

    orders_xml = table_picture.findall("Orders/Order")
    for order_xml in orders_xml:
        if order_xml.get("state") not in ["IN_PROGRESS", "TOTALED"]:
            continue

        user_id_from_session = get_user_id_from_session(session_id=order_xml.get("sessionId"))

        is_same_user = user_id_from_session == user_id
        if is_same_user and table.current_pos_id is not None:
            doStoreOrder(
                pos_id=table.current_pos_id,
                pos_id_to_show_message=pos_id,
                show_popup=False,
            )
            return True

    return False


def _voided_quick_service_order_after_confirmation(
    pos_id,                 # type: str
    pos_to_close_user,      # type: str
    order_id=None,              # type: Optional[str]
):
    # type: (...) -> bool

    confirmation = _get_confirmation_close_user_and_voided_order(pos_id=pos_id)
    if confirmation == 0:
        do_void_order(
            pos_id=pos_to_close_user,
            received_void_reason="CANCELLATION",
            show_cancel_message=False,
            order_id=order_id,
        )
        return True

    return False


def _voided_table_order_after_confirmation(
    pos_id,                 # type: str
    pos_to_close_user,      # type: str
    order_id,               # type: str
    table_id,               # type: str
):
    # type: (...) -> bool

    confirmation = _get_confirmation_close_user_and_voided_order(pos_id=pos_id)
    if confirmation == 0:
        from table_actions import void_table_order_without_void_reason

        return void_table_order_without_void_reason(
            pos_id=pos_to_close_user,
            order_id=order_id,
            table_id=table_id,
        )

    return False


def get_order_picture(
    pos_id,         # type: str
    order_id=None,  # type: str
):
    # type: (Union[str, int], Optional[str]) -> Optional[Element]

    data = "{}\0{}".format(str(pos_id), str(order_id or ''))
    reply = sysactions.send_message(
        dest_name="ORDERMGR{}".format(str(pos_id)),
        token=TK_SLCTRL_OMGR_ORDERPICT,
        format=FM_PARAM,
        data=data,
    )

    if reply.token == TK_SYS_ACK:
        try:
            split_reply = reply.data.split("\0")
            if not split_reply[2]:
                return None
            order_picture = eTree.XML(split_reply[2])
            if order_id:
                order_picture = order_picture.findall('.//Order')[0]
            return order_picture
        except AttributeError:
            return None

    raise Exception("Error getting OrderPicture: {} - {}".format(reply.token, reply.data))


def common_close_user(
    pos_id,                     # type: str
    model,                      # type: eTree.XML
    user_id,                    # type: str
    authorizer=None,            # type: Optional[str]
    pos_to_close_user=None,     # type: Optional[str]
):
    # type: (...) -> bool

    if not acquire_user_lock_close_operator(user_id=user_id):
        sysactions.show_messagebox(
            posid=pos_id,
            message=get_lock_message(user_id=user_id),
            icon="warning",
        )
        raise sysactions.StopAction()

    try:
        return _inner_common_close_user(
            pos_id=pos_id,
            model=model,
            user_id=user_id,
            authorizer=authorizer,
            pos_to_close_user=pos_to_close_user,
        )
    finally:
        release_user_lock(user_id=user_id)


def _inner_common_close_user(
    pos_id,                     # type: str
    model,                      # type: eTree.XML
    user_id,                    # type: str
    authorizer=None,            # type: Optional[str]
    pos_to_close_user=None,     # type: Optional[str]
):
    # type: (...) -> bool

    from table_actions import considers_tef_on_bordereau

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    is_logged_in, session = get_operator_info(
        model=model,
        operator_id=user_id,
    )
    if not session:
        sysactions.show_messagebox(pos_id, "$USER_NOT_OPENED", "$ERROR")
        raise NotFoundUserError()

    pod_function = sysactions.get_posfunction(model=model)
    current_order = sysactions.get_current_order(model=model)

    pod_type = sysactions.get_cfg(pos_id).key_value("pod") or "FL"
    is_quick_service = pod_type in ["FC", "FL"]

    if current_order is not None:
        table_id = current_order.find("CustomOrderProperties/OrderProperty/[@key='TABLE_ID']")

        try:
            order = sysactions.get_current_order(model=model)

            if table_id is None:
                close_quick_service_order_from_another_pos(
                    pos_id=pos_id,
                    order=order,
                    user_id=user_id,
                    pos_to_close_user=pos_to_close_user,
                )
            else:
                close_table_service_from_another_pos(
                    pos_id=pos_id,
                    order=order,
                    user_id=user_id,
                    pos_to_close_user=pos_to_close_user,
                )
        except (CloseOrderByOperatorCloseException, Exception) as ex:
            logger.info("Operation not complete because : {}".format(ex.message))
            raise sysactions.StopAction()

        if is_logged_in:
            pause_user(
                pos_id=str(pos_to_close_user),
                pos_id_to_show_message=pos_id,
            )

    period = sysactions.get_business_period(model=model)

    transfer_values = {}
    if pod_function != "OT":
        transfer_values = get_transfer_values(
            pos_id=pos_id,
            period=period,
            session=session,
        )

    report_name = "close_operator_report"

    bordereau_enabled = False
    if pod_function != "OT":
        bordereau_enabled = get_close_operator_conference_type() == ConferenceTypes.BORDEREAU.value
        response = do_transfer(
            pos_id=pos_id,
            transfer_type=6,
            operator_id=user_id,
            authorizer=authorizer,
            pos_to_close_user=pos_to_close_user,
            transfer_values=transfer_values,
        )
        set_authorizer(pos_id, authorizer, is_quick_service)

        if response['need_ask_justification']:
            sysactions.print_report(
                pos_id,
                None,
                True,
                report_name,
                pos_to_close_user,
                period,
                user_id,
                session,
                authorizer,
                bordereau_enabled,
                considers_tef_on_bordereau,
            )

            account_repository = AccountRepository(mb_context=mb_context)
            get_justification(
                account_repository=account_repository,
                pos_id=pos_id,
                session=session,
            )

    if close_user(
        pos_id=pos_id,
        user_id=user_id,
        pos_to_close_user=pos_to_close_user,
    ):
        set_authorizer(
            pos_id=pos_id,
            authorizer=authorizer,
            is_quick_service=is_quick_service
        )
        logger.debug("Logoff POS %d - Operador %s deslogado com sucesso" % (int(pos_to_close_user), user_id))
    else:
        logger.error("Logoff POS %d - Falha ao deslogar operador %s" % (int(pos_to_close_user), user_id))
        sysactions.show_messagebox(pos_id, "$OPERATION_FAILED", "$ERROR")
        raise sysactions.StopAction()

    logger.debug("Logoff POS %d - Deslogando operador %s" % (int(pos_id), user_id))

    sysactions.print_report(
        pos_id,
        None,
        True,
        report_name,
        pos_to_close_user,
        period,
        user_id,
        session,
        authorizer,
        bordereau_enabled,
        considers_tef_on_bordereau,
    )

    return True


def set_authorizer(pos_id, authorizer, is_quick_service):
    if authorizer is not None and is_quick_service is False:
        sysactions.set_custom(pos_id, 'Last Manager ID', authorizer, True)
        sysactions.set_custom(pos_id, 'Authorization Level', get_user_level(authorizer), True)


def _get_confirmation_close_user_and_voided_order(
    pos_id,     # type: str
):
    # type: (...) -> Union[None, int]

    return sysactions.show_messagebox(
        posid=pos_id,
        message="$CLOSE_USER_VOID_ORDER",
        timeout=720000,
        buttons="$OK|$CANCEL",
    )



from production.box import OrderStateFilterBox
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class OrderStateFilterBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(OrderStateFilterBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> OrderStateFilterBox
        return OrderStateFilterBox(config.name,
                                   config.sons,
                                   self.get_extra(config, "AllowedOrderStates", []),
                                   self.get_extra(config, "ExcludedOrderStates", []),
                                   self.get_logger(config))



import time
import uuid
from logging import Logger
from xml.etree import cElementTree as eTree
from threading import (
    Lock,
    Thread,
)
from typing import (
    List,
    Optional,
)

import cfgtools
from xml.etree.ElementTree import Element

from messagehandler import MessageHandler
from messagebus import (
    AckMessage,
    DefaultToken,
    Event,
    Message,
    MessageBus,
    MessageBusException,
    NakMessage,
    TokenCreator,
    TokenPriority,
)
from msgbus import (
    FM_PARAM,
    MBEasyContext
)
from model import (
    Kds,
    KdsModel,
    KdsView,
    PickupConfig,
)

from services import DeliveryService

MSG_GRP_KDS = '0B'
TK_KDS_GETMODEL = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "1")
TK_KDS_UPDATE_VIEW = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "2")
TK_KDS_REFRESH = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "3")
TK_KDS_REFRESH_END = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "4")
TK_KDS_SET_PROD_STATE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "5")
TK_KDS_UNDO_SERVE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "6")
TK_KDS_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "8")
TK_KDS_CHANGE_VIEW = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "A")
TK_KDS_BUMP_BAR_COMMAND = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "B")
TK_KDS_BUMP_ZOOM_NEXT = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "C")
TK_KDS_RELOAD = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "14")
TK_KDS_SELECT_ORDER = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "20")
TK_KDS_UPDATE_STATISTICS = TokenCreator.create_token(TokenPriority.high, MSG_GRP_KDS, "21")

MSG_GRP_PRODUCTION = "0C"
TK_PROD_REFRESH_VIEW = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "1")
TK_PROD_SET_ORDER_STATE = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "2")
TK_PROD_UNDO = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "7")
TK_PROD_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "8")
TK_PROD_GET_STATISTICS = TokenCreator.create_token(TokenPriority.low, MSG_GRP_PRODUCTION, "102")

NEW_PRODUCTION_COMP = "ProductionSystem"
NO_KDS_FOUND_MESSAGE = "No KDS found"


kds_list_lock = Lock()


class ProxyMessageHandler(MessageHandler):
    def __init__(
        self,
        comp_name,    # type: str
        mb_context,   # type: MBEasyContext
        message_bus,  # type: MessageBus
        config,       # type: cfgtools.Group
        logger        # type: Logger
    ):
        # type: (...) -> None

        self.comp_name = comp_name
        self.kds_list = None
        self.brands_loaded = False
        self.mb_context = mb_context
        self.message_bus = message_bus
        self.logger = logger
        self.commands_uuid = []
        self.config = config

        self.statistics_cache = ''
        init_thread = Thread(target=self.get_current_statistics)
        init_thread.daemon = True
        init_thread.start()

    def handle_event(self, message_bus, event):
        return

    def handle_message(self, message_bus, message):
        data = message.data.split('\0')
        resp = None

        if message.token == TK_KDS_SELECT_ORDER:
            kds_id = int(data[0])
            self.logger.info("Receiving TK_KDS_SELECT_ORDER for KDS - {:02d}".format(kds_id))

            # TODO: Change Selected Order and Store it
            message.token = DefaultToken.TK_SYS_ACK

        elif message.token == TK_KDS_GETMODEL:
            resp = self.handle_kds_get_model(data)
        elif message.token == TK_KDS_CHANGE_VIEW:
            resp = self.handle_kds_change_view(data)
        elif message.token == TK_KDS_UPDATE_VIEW:
            resp = self.handle_kds_update_view(data, message_bus)
        elif message.token == TK_KDS_REFRESH:
            resp = self.handle_kds_refresh(data, message_bus)
        elif message.token == TK_KDS_REFRESH_END:
            resp = self.handle_kds_refresh_end(data, message_bus)
        elif message.token == TK_KDS_SET_PROD_STATE:
            resp = self.handle_kds_set_prod_state(data, message_bus)
        elif message.token == TK_KDS_UNDO_SERVE:
            resp = self.handle_kds_undo_serve(data, message_bus)
        elif message.token == TK_KDS_UPDATE_STATISTICS:
            resp = self.handle_kds_update_statistics(data, message_bus)
        elif message.token == TK_KDS_BUMP_BAR_COMMAND:
            resp = self.handle_kds_bump_bar_command(data, message_bus)
        elif message.token == TK_KDS_BUMP_ZOOM_NEXT:
            resp = self.handle_kds_zoom_next(data)
        elif message.token == TK_KDS_TOGGLE_TAG_LINE:
            resp = self.handle_kds_toggle_tag_line(data, message_bus)
        elif message.token == TK_KDS_RELOAD:
            resp = self.handle_kds_reload(data, message_bus)
        else:
            resp = self.forward_message(message, message_bus)

        message_bus.reply_message(message, resp)
        self.logger.info("Message replied")

    def handle_kds_get_model(self, data):
        self.logger.info("TK_KDS_GETMODEL - Received")

        kds_id = data[0]
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_GETMODEL - KdsId: {} not found".format(kds_id))
            return NakMessage(NO_KDS_FOUND_MESSAGE, 0)

        self.logger.info("TK_KDS_GETMODEL - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        kds_model_xml.append(eTree.Element("Statistics", text=self.statistics_cache))
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_GETMODEL - Finished")
        return resp

    def handle_kds_change_view(self, data):
        self.logger.info("TK_KDS_CHANGE_VIEW - Received")

        kds_id = data[0]
        view_name = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_CHANGE_VIEW - KdsId: {} not found".format(kds_id))
            return NakMessage(NO_KDS_FOUND_MESSAGE, 0)

        if self._is_uuid_unique(cmd_uuid):
            self.set_selected_view(kds, view_name)
            self.logger.info("TK_KDS_CHANGE_VIEW - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_CHANGE_VIEW - Finished")
        return resp

    def get_kds(self, kds_id):
        return next((x for x in self._get_kds_list(need_brands=True) if x.name == kds_id), None)

    @staticmethod
    def get_bump_bar_command(bump_bar_commands, command_code):
        return next((x for x in bump_bar_commands if command_code in x.keys), None)

    @staticmethod
    def set_selected_view(kds, view_name):
        for view in kds.views:
            view.selected = view.name == view_name

        has_selected_view = next((x for x in kds.views if x.selected), None)
        if not has_selected_view:
            kds.views[0].selected = True

    @staticmethod
    def set_view_zoom(kds, production_view):
        view = next((x for x in kds.views if x.production_view == production_view), None)
        if not view:
            return

        view.selected_zoom += 1
        if view.selected_zoom == len(view.zoom_levels):
            view.selected_zoom = 0

    def handle_kds_update_view(self, data, message_bus):
        self.logger.info("TK_KDS_UPDATE_VIEW - Received")

        production_view = data[0]
        xml = eTree.XML(data[1])
        order_id = xml.get("order_id")

        self.logger.info("TK_KDS_UPDATE_VIEW - ProductionView: {} / OrderId: {}".format(production_view, order_id))

        view_update_xml = self.create_view_update_xml(production_view, xml)
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_UPDATE_VIEW - Sending event for KDS: {} / OrderId: {}".format(kds, order_id))

            event_subject = "KDS" + kds
            event_type = "PROD_ORDER_" + order_id + "_" + production_view
            event = self.send_event(event_subject, event_type, view_update_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_UPDATE_VIEW - Finished")

        return AckMessage('', 0)

    def handle_kds_refresh(self, data, message_bus):
        self.logger.info("TK_KDS_REFRESH - Received")
        response = AckMessage('', 0)

        kds_id = data[0]
        for view in self.get_production_views_with_kds_id(kds_id):
            self.logger.info("TK_KDS_REFRESH - Sending event for View: {}".format(view))

            message = Message(TK_PROD_REFRESH_VIEW, FM_PARAM, str(view), 3000000)
            reply = message_bus.send_message(NEW_PRODUCTION_COMP, message)
            if reply.token != DefaultToken.TK_SYS_ACK:
                response = NakMessage()

        self.logger.info("TK_KDS_REFRESH - Finished")
        return response

    def handle_kds_set_prod_state(self, data, message_bus):
        self.logger.info("TK_KDS_SET_PROD_STATE - Received")

        kds_id = data[0]
        order_id = data[1]
        prod_state = data[2]
        view = data[3]
        pickup_area_info = data[4] if len(data) > 4 else ''
        cmd_uuid = data[5] if len(data) > 5 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_SET_PROD_STATE - Sending event to KDS: {} / OrderId: {} / ProdState {}"
            self.logger.info(log_message.format(kds_id, order_id, prod_state))

            message_data = order_id + "\0" + view + "\0" + prod_state + "\0" + pickup_area_info
            message = Message(TK_PROD_SET_ORDER_STATE, FM_PARAM, message_data, 3000000)
            message_bus.send_message(NEW_PRODUCTION_COMP, message)

        self.logger.info("TK_KDS_SET_PROD_STATE - Finished")
        return AckMessage('', 0)

    def handle_kds_zoom_next(self, data):
        self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - Received")

        kds_id = data[0]
        production_view = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'
        kds = self.get_kds(kds_id)

        if not kds:
            self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - KdsId: {} not found".format(kds_id))
            return NakMessage(NO_KDS_FOUND_MESSAGE, 0)

        if self._is_uuid_unique(cmd_uuid):
            self.set_view_zoom(kds, production_view)
            self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - KdsId: {}".format(kds_id))

        kds_model = KdsModel(kds, kds_id)
        kds_model_xml = kds_model.get_model_xml()
        resp = AckMessage(eTree.tostring(kds_model_xml, encoding='utf8', method='xml'), 0)

        self.logger.info("TK_KDS_BUMP_ZOOM_NEXT - Finished")
        return resp

    def handle_kds_refresh_end(self, data, message_bus):
        self.logger.info("TK_KDS_REFRESH_END - Received")

        production_view = data[0]
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_REFRESH_END - Sending event to KDS: {} - VIEW: {}".format(kds, production_view))

            event_subject = "KDS" + kds
            event_type = "REFRESH_END_" + production_view
            refresh_end_xml = eTree.XML("<RefreshEnd name=\"{}\" updated=\"true\"/>".format(production_view))
            event = self.send_event(event_subject, event_type, refresh_end_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_REFRESH_END - Finished")

        return AckMessage('', 0)

    def handle_kds_reload(self, data, message_bus):
        self.logger.info("TK_KDS_RELOAD - Received")

        production_view = data[0]
        for kds in self.get_kds_list_with_production_view(production_view):
            self.logger.info("TK_KDS_RELOAD - Sending event to KDS: {}".format(kds))

            event_subject = "KDS" + kds
            event_type = "RELOAD_" + production_view
            refresh_end_xml = eTree.XML("<Reload name=\"{}\"/>".format(production_view))
            event = self.send_event(event_subject, event_type, refresh_end_xml)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_RELOAD - Finished")

        return AckMessage('', 0)

    def handle_kds_toggle_tag_line(self, data, message_bus):
        self.logger.info("TK_KDS_TOGGLE_TAG_LINE - Received")

        kds_id = data[0]
        order_id = data[1]
        line_number = data[2]
        tag_name = data[3]
        view = data[4]
        cmd_uuid = data[5] if len(data) > 5 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_TOGGLE_TAG_LINE - Sending event to KDS: {} / OrderId: {} / LineNumber {} / TagName {}"
            self.logger.info(log_message.format(kds_id, order_id, line_number, tag_name))

            message_data = order_id + "\0" + view + "\0" + line_number + "\0" + tag_name
            message = Message(TK_PROD_TOGGLE_TAG_LINE, FM_PARAM, message_data, 3000000)
            message_bus.send_message(NEW_PRODUCTION_COMP, message)

        self.logger.info("TK_KDS_TOGGLE_TAG_LINE - Finished")
        return AckMessage('', 0)

    def handle_kds_update_statistics(self, data, message_bus):
        self.logger.info("TK_KDS_UPDATE_STATISTICS - Received")

        try:
            statistics = data[0]
            self.statistics_cache = statistics
        except ValueError:
            pass

        for kds_id in self.get_all_kds_ids():
            event_data = eTree.tostring(self.create_statistics_event_from_json(self.statistics_cache))
            event = Event(subject='KDS{}'.format(kds_id), event_type='Statistics', data=event_data)
            message_bus.publish_event(event)

        self.logger.info("TK_KDS_UPDATE_STATISTICS - Finished")

        return AckMessage('', 0)

    def handle_kds_bump_bar_command(self, data, message_bus):
        self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Received")

        kds_id = data[0]
        bump_bar_name = data[1]
        command_code = data[2]

        log_message = "TK_KDS_BUMP_BAR_COMMAND - KdsId: {} / BumpBar: {} / Command: {}"
        self.logger.info(log_message.format(kds_id, bump_bar_name, command_code))

        kds = self.get_kds(kds_id)
        if not kds:
            self.logger.info("TK_KDS_BUMP_BAR_COMMAND - KdsId: {} not found".format(kds_id))
            return NakMessage(NO_KDS_FOUND_MESSAGE, 0)

        bump_bar = kds.bump_bar
        command = self.get_bump_bar_command(bump_bar.commands, command_code)
        if not command:
            self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Key not mapped")
            return AckMessage('', 0)

        root_element = eTree.Element("Event")
        command_uuid = str(uuid.uuid4())
        eTree.SubElement(root_element, "Command", name=command.name, value=command_uuid)
        event = Event(subject='KDS{}'.format(kds_id), event_type='KdsModel', data=eTree.tostring(root_element))
        message_bus.publish_event(event)

        self.logger.info("TK_KDS_BUMP_BAR_COMMAND - Received - UUID: {}".format(command_uuid))
        return AckMessage('', 0)

    def handle_kds_undo_serve(self, data, message_bus):
        self.logger.info("TK_KDS_UNDO_SERVE - Received")

        kds_id = data[0]
        view = data[1]
        cmd_uuid = data[2] if len(data) > 2 else '0'

        if self._is_uuid_unique(cmd_uuid):
            log_message = "TK_KDS_UNDO_SERVE - Sending event to KDS: {}"
            self.logger.info(log_message.format(kds_id))

            message = Message(TK_PROD_UNDO, FM_PARAM, str(view), 3000000)
            message_bus.send_message(NEW_PRODUCTION_COMP, message)

        self.logger.info("TK_KDS_UNDO_SERVE - Finished")
        return AckMessage('', 0)

    @staticmethod
    def create_view_update_xml(production_view, xml):
        view_update_xml = eTree.Element("ViewUpdate", name=production_view)
        view_update_xml.append(xml)

        return view_update_xml

    @staticmethod
    def send_event(event_subject, event_type, xml):
        event_xml = eTree.Element("Event", subject=event_subject, type=event_type)
        event_xml.append(xml)
        event = Event(event_subject, event_type, eTree.tostring(event_xml))

        return event

    def forward_message(self, message, _):
        resp = AckMessage('', 0)
        try:
            self.logger.info("Receiving message for KDS - data = {}".format(message))
            self.logger.info("Message Ignored")
        except MessageBusException:
            self.logger.info("Error forwarding message - Component not found")

        return resp

    def get_kds_list_with_production_view(self, kds_view):
        kds_list = []
        for kds in self._get_kds_list(need_brands=False):
            if len([v for v in kds.views if v.production_view == kds_view]) > 0:
                kds_list.append(kds.name)

        return list(set(kds_list))

    def get_production_views_with_kds_id(self, kds_id):
        view_list = []

        kds = next((x for x in self._get_kds_list(need_brands=False) if x.name == kds_id), None)
        for view in kds.views:
            view_list.append(view.production_view)

        return list(set(view_list))

    @staticmethod
    def create_statistics_event_from_json(json_data):
        # type: (str) -> Element
        root_element = eTree.Element("Event")
        eTree.SubElement(root_element, "Statistics", text=json_data)
        return root_element

    def get_all_kds_ids(self):
        kds_list = []
        for kds in self._get_kds_list(need_brands=False):
            kds_list.append(kds.name)

        return list(set(kds_list))

    def get_current_statistics(self):
        statistics_ok = False
        while statistics_ok is False:
            try:
                resp = self.message_bus.send_message(NEW_PRODUCTION_COMP, Message(TK_PROD_GET_STATISTICS))
                if resp.token == DefaultToken.TK_SYS_ACK:
                    self.statistics_cache = resp.data
                    statistics_ok = True
            except (BaseException,):
                time.sleep(0.500)

    def _is_uuid_unique(self, cmd_uuid):
        # type: (str) -> bool

        self.logger.info('UUID: {}'.format(cmd_uuid))
        if cmd_uuid != '0' and cmd_uuid in self.commands_uuid:
            self.logger.info('Command Ignored')
            return False

        self.commands_uuid.append(cmd_uuid)
        self.commands_uuid = self.commands_uuid[-5:]
        return True

    def _get_kds_list(
            self,           # type: ProxyMessageHandler
            need_brands     # type: bool
    ):
        # type: (...) -> List[Kds]

        with kds_list_lock:
            return self._inner_get_kds_list(need_brands=need_brands)

    def _inner_get_kds_list(
            self,           # type: ProxyMessageHandler
            need_brands,    # type: bool
    ):
        # type: (...) -> Optional[List[Kds]]

        if self.kds_list and self.brands_loaded:
            return self.kds_list

        if self.kds_list is None:
            kds_views = self._get_kds_views()
            kds_list = []
            kds_config = self.config.find_group("KDS")
            if not kds_config:
                return []

            for kds_config_group in kds_config.groups:
                kds_list.append(
                    Kds(
                        kds_config=kds_config_group,
                        kds_views=kds_views,
                        brands=None,
                        pickup_config=self._get_pickup_config(),
                    )
                )

            self.kds_list = kds_list

        if need_brands:
            delivery_service = DeliveryService(self.mb_context, self.logger)
            delivery_brands = delivery_service.get_delivery_brands()
            if delivery_brands is None:
                return self.kds_list

            for kds in self.kds_list:
                kds.set_brands(delivery_brands)

            self.brands_loaded = True

        return self.kds_list

    def _get_kds_views(self):
        # type: (...) -> List[KdsView]

        views = []
        view_config = self.config.find_group("Views")
        if not view_config:
            return []

        for view_config_group in view_config.groups:
            views.append(
                KdsView(view_config_group)
            )

        return views

    def _get_pickup_config(
        self,   # type: ProxyMessageHandler
    ):
        # type: (...) -> PickupConfig

        pickup_config = self.config.find_group("PickupConfig")
        if not pickup_config:
            return []

        return PickupConfig(pickup_config.groups[0])



from logging import Logger

from production.box import ProductionBox
from production.command import ProductionCommandProcessor
from production.manager import PublishScheduler, OrderChanger, StatisticUpdater
from production.repository import ProductRepository
from production.treebuilder import BoxConfiguration
from production.treebuilder.boxfactory.builder import (
    OrderStateFilterBoxBuilder,
    OrderSubTypeFilterBoxBuilder,
    JitItemFilterBoxBuilder,
    HideOptionBoxBuilder,
    AllDoneBlinkBoxBuilder,
    ComboDoneWhenAllDoneBoxBuilder,
    OrderSequencerBoxBuilder,
    RemoveDoingWhenDoneBoxBuilder,
    OrderStateTimerBoxBuilder,
    FromOrderStateFilterBoxBuilder,
    CustomParamCourseBoxBuilder,
    CustomParamPrepTimeBoxBuilder,
    ShowOnKitchenFilterBoxBuilder,
    NoJitAddTagBoxBuilder,
    ItemPropertiesFilterBoxBuilder,
    AllServedItemsBoxBuilder,
    SendMomentTimerBuilder,
    ConditionBoxBuilder,
    CFHHiderBoxBuilder,
    CustomPropertyFilterBoxBuilder,
    ComboFilterBoxBuilder,
    DistributorBoxBuilder,
    OrderRoundRobinBoxBuilder,
    RemoveDeletedItemsBoxBuilder,
    FilterDeletedItemsBeforeFirstStateBoxBuilder,
    OrderSpreaderBoxBuilder,
    OrderStampBoxBuilder,
    OrderMergerBoxBuilder,
    RecalledOrderStampBoxBuilder,
    ComboServedWhenAllServedBoxBuilder,
    FilterOrderIfNotAllItemsHaveTagBoxBuilder,
    AllItemsDoneBoxBuilder,
    RemoveReprocessedOrdersBoxBuilder,
    ProductCodeFilterBoxBuilder,
    OrderProductCategoryFilterBoxBuilder,
    PodTypeFilterBoxBuilder,
    SaleTypeFilterBoxBuilder,
    SkipCourseBoxBuilder,
    CollapseSameItemsBoxBuilder,
    StampItemsViewBoxBuilder,
    ItemSequencerBoxBuilder,
    StatisticBoxBuilder,
    AddTimeTagBoxBuilder,
    ProdStateFilterBoxBuilder,
    CustomPropertyValueFilterBoxBuilder,
    ItemSorterBoxBuilder,
    DisplayTimeFilterBoxBuilder,
    ProdStateTimerBoxBuilder,
    DecreaseTimeTagBoxBuilder,
    AllItemHasAtLeastOneTagFilterBoxBuilder,
    FilterOrderIfAllItemsHaveTagBoxBuilder,
    FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder,
    ComboTaggedWhenAllHasTagBoxBuilder,
    SorterBoxBuilder,
    TagItemsWhenProdStateBoxBuilder,
    FilterDefaultQuantityIngredientsBoxBuilder,
    JitAddTagBoxBuilder,
    ReprocessOrdersBoxBuilder,
    JitSequencerItemFilterBoxBuilder,
    SequencerTagItemBoxBuilder,
    OrderSplitterBoxBuilder,
    OrderPropertiesFilterBoxBuilder,
    NotShowOnKitchenAddTagBoxBuilder,
    OrderUnsplitterBoxBuilder,
    FilterDeletedItemsNotSentToKitchenBoxBuilder,
    NonComboFilterBoxBuilder,
    ItemCustomParamsFilterBoxBuilder,
    NonValidJitAddTagBoxBuilder,
)

from production.view import ProductionView
from typing import Dict

from .builder import TagFilterBoxBuilder, ViewBoxBuilder


class ProductionBoxFactory(object):

    def __init__(
        self,                   # type: ProductionBoxFactory
        command_processor,      # type: ProductionCommandProcessor
        publish_scheduler,      # type: PublishScheduler
        order_changer,          # type: OrderChanger
        statistic_updater,      # type: StatisticUpdater
        product_repository,     # type: ProductRepository
        loggers,                # type: Dict[unicode, Logger]
    ):
        # type: (...) -> None

        self.command_processor = command_processor
        self.publish_scheduler = publish_scheduler
        self.order_changer = order_changer
        self.product_repository = product_repository
        self.loggers = loggers

        self.builders = {
            u"TagFilterBox": TagFilterBoxBuilder(loggers),
            u"OrderStateFilterBox": OrderStateFilterBoxBuilder(product_repository, loggers),
            u"OrderSubTypeFilterBox": OrderSubTypeFilterBoxBuilder(product_repository, loggers),
            u"JitItemFilterBox": JitItemFilterBoxBuilder(product_repository, loggers),
            u"HideOptionBox": HideOptionBoxBuilder(product_repository, loggers),
            u"ComboDoneWhenAllDoneBox": ComboDoneWhenAllDoneBoxBuilder(loggers),
            u"ComboServedWhenAllServedBox": ComboServedWhenAllServedBoxBuilder(loggers),
            u"AllDoneBlinkBox": AllDoneBlinkBoxBuilder(loggers),
            u"OrderSequencerBox": OrderSequencerBoxBuilder(product_repository, loggers),
            u"RemoveDoingWhenDoneBox": RemoveDoingWhenDoneBoxBuilder(loggers),
            u"OrderStateTimerBox": OrderStateTimerBoxBuilder(loggers),
            u"FromOrderStateFilterBox": FromOrderStateFilterBoxBuilder(product_repository, loggers),
            u"CustomParamCourseBox": CustomParamCourseBoxBuilder(product_repository, publish_scheduler, loggers),
            u"CFHHiderBox": CFHHiderBoxBuilder(product_repository, loggers),
            u"ComboFilterBox": ComboFilterBoxBuilder(
                product_repository=product_repository,
                loggers=loggers,
            ),
            u"DistributorBox": DistributorBoxBuilder(loggers),
            u"CustomParamPrepTimeBox": CustomParamPrepTimeBoxBuilder(product_repository, publish_scheduler, loggers),
            u"ShowOnKitchenFilterBox": ShowOnKitchenFilterBoxBuilder(
                product_repository=product_repository,
                loggers=loggers
            ),
            u"NotShowOnKitchenAddTagBox": NotShowOnKitchenAddTagBoxBuilder(
                product_repository=product_repository,
                loggers=loggers
            ),
            u"NoJitAddTagBox": NoJitAddTagBoxBuilder(product_repository, loggers),
            u"ItemPropertiesFilterBox": ItemPropertiesFilterBoxBuilder(loggers),
            u"AllServedItemsBox": AllServedItemsBoxBuilder(loggers),
            u"SendMomentTimerBox": SendMomentTimerBuilder(loggers),
            u"ConditionBox": ConditionBoxBuilder(loggers),
            u"CustomPropertyFilterBox": CustomPropertyFilterBoxBuilder(loggers),
            u"OrderRoundRobinBox": OrderRoundRobinBoxBuilder(order_changer, loggers),
            u"RemoveDeletedItemsBox": RemoveDeletedItemsBoxBuilder(loggers),
            u"FilterDeletedItemsBeforeFirstStateBox": FilterDeletedItemsBeforeFirstStateBoxBuilder(loggers),
            u"OrderSpreaderBox": OrderSpreaderBoxBuilder(loggers),
            u"OrderStampBox": OrderStampBoxBuilder(loggers),
            u"OrderMergerBox": OrderMergerBoxBuilder(loggers),
            u"RecalledOrderStampBox": RecalledOrderStampBoxBuilder(loggers),
            u"FilterOrderIfNotAllItemsHaveTagBox": FilterOrderIfNotAllItemsHaveTagBoxBuilder(loggers),
            u"AllItemsDoneBox": AllItemsDoneBoxBuilder(loggers),
            u"RemoveReprocessedOrdersBox": RemoveReprocessedOrdersBoxBuilder(
                command_processor=command_processor,
                loggers=loggers,
            ),
            u"ProductCodeFilterBox": ProductCodeFilterBoxBuilder(loggers),
            u"OrderProductCategoryFilterBox": OrderProductCategoryFilterBoxBuilder(loggers),
            u"PodTypeFilterBox": PodTypeFilterBoxBuilder(loggers),
            u"SaleTypeFilterBox": SaleTypeFilterBoxBuilder(loggers),
            u"SkipCourseBox": SkipCourseBoxBuilder(loggers),
            u"CollapseSameItemsBox": CollapseSameItemsBoxBuilder(loggers),
            u"ItemSequencerBox": ItemSequencerBoxBuilder(order_changer, loggers),
            u"StatisticBox": StatisticBoxBuilder(statistic_updater, loggers),
            u"AddTimeTagBox": AddTimeTagBoxBuilder(loggers),
            u"ProdStateFilterBox": ProdStateFilterBoxBuilder(product_repository, loggers),
            u"CustomPropertyValueFilterBox": CustomPropertyValueFilterBoxBuilder(loggers),
            u"ItemSorterBox": ItemSorterBoxBuilder(product_repository, loggers),
            u"SorterBox": SorterBoxBuilder(product_repository, loggers),
            u"DisplayTimeFilterBox": DisplayTimeFilterBoxBuilder(loggers),
            u"ProdStateTimerBox": ProdStateTimerBoxBuilder(loggers),
            u"DecreaseTimeTagBox": DecreaseTimeTagBoxBuilder(loggers),
            u"AllItemHasAtLeastOneTagFilterBox": AllItemHasAtLeastOneTagFilterBoxBuilder(loggers),
            u"FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBox": FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder(loggers),  # noqa
            u"ComboTaggedWhenAllHasTagBox": ComboTaggedWhenAllHasTagBoxBuilder(loggers),
            u"TagItemsWhenProdStateBox": TagItemsWhenProdStateBoxBuilder(loggers),
            u"FilterOrderIfAllItemsHaveTagBox": FilterOrderIfAllItemsHaveTagBoxBuilder(loggers),
            u"FilterDefaultQuantityIngredientsBox": FilterDefaultQuantityIngredientsBoxBuilder(loggers),
            u"JitAddTagBox": JitAddTagBoxBuilder(product_repository, loggers),
            u"ReprocessOrdersBox": ReprocessOrdersBoxBuilder(publish_scheduler, loggers),
            u"JitSequencerItemFilterBox": JitSequencerItemFilterBoxBuilder(product_repository, loggers),
            u"SequencerTagItemBox": SequencerTagItemBoxBuilder(loggers),
            u"OrderSplitterBox": OrderSplitterBoxBuilder(order_changer, publish_scheduler, loggers),
            u"OrderPropertiesFilterBox": OrderPropertiesFilterBoxBuilder(order_changer, loggers),
            u"OrderUnsplitterBox": OrderUnsplitterBoxBuilder(loggers=loggers),
            u"FilterDeletedItemsNotSentToKitchenBox": FilterDeletedItemsNotSentToKitchenBoxBuilder(loggers=loggers),
            u"NonComboFilterBox": NonComboFilterBoxBuilder(loggers=loggers),
            u"ItemCustomParamsFilterBox": ItemCustomParamsFilterBoxBuilder(
                product_repository=self.product_repository,
                loggers=loggers
            ),
            u"NonValidJitAddTagBox": NonValidJitAddTagBoxBuilder(loggers=loggers)
        }

    def set_views(self, views):
        # type: (Dict[str, ProductionView]) -> None
        self.builders[u"ViewBox"] = ViewBoxBuilder(self.command_processor, self.order_changer, views, self.loggers)
        self.builders[u"StampItemsView"] = StampItemsViewBoxBuilder(self.order_changer, views, self.loggers)

    def build(self, config):
        # type: (BoxConfiguration) -> ProductionBox
        if config.type not in self.builders:
            raise Exception("Builder not found for box: {} - {}".format(config.name, config.type))

        return self.builders[config.type].build(config)
